import numpy as np
import torch
import pytest
from goldener.torch_utils import (
    torch_tensor_to_numpy_vectors,
    numpy_vectors_to_torch_tensor,
    np_transform_from_torch,
    make_2d_tensor,
    ResetableTorchIterableDataset,
)


class TestTensorToNumpyVectors:
    def test_torch_tensor_to_numpy_vectors_0d(self):
        t = torch.tensor(3)
        arr = torch_tensor_to_numpy_vectors(t)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (1, 1)

    def test_torch_tensor_to_numpy_vectors_1d(self):
        t = torch.rand(3)
        arr = torch_tensor_to_numpy_vectors(t)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (3, 1)

    def test_torch_tensor_to_numpy_vectors_2d(self):
        t = torch.rand(3, 2)
        arr = torch_tensor_to_numpy_vectors(t)
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (3, 2)

    def test_torch_tensor_to_numpy_vectors_3d(self):
        t = torch.rand(2, 3, 4, dtype=torch.float32)
        arr = torch_tensor_to_numpy_vectors(t)
        assert arr.shape == (
            8,
            3,
        )


class TestNumpyVectorToTensor:
    def test_numpy_vectors_to_torch_tensor_0d(self):
        arr = np.array(3)
        shape = (1, 1)
        t = numpy_vectors_to_torch_tensor(
            arr, shape, torch.float32, torch.device("cpu")
        )
        assert isinstance(t, torch.Tensor)
        assert t.shape == shape
        assert t.device.type == "cpu"
        assert t.dtype == torch.float32

    def test_numpy_vectors_to_torch_tensor_1d(self):
        arr = np.random.randint(0, 10, size=(3,), dtype=np.int32)
        shape = (3, 1)
        t = numpy_vectors_to_torch_tensor(
            arr, shape, torch.float32, torch.device("cpu")
        )
        assert isinstance(t, torch.Tensor)
        assert t.shape == shape
        assert t.device.type == "cpu"
        assert t.dtype == torch.float32

    def test_numpy_vectors_to_torch_tensor_2d(self):
        arr = np.random.randint(0, 10, size=(3, 2), dtype=np.int32)
        shape = (3, 2)
        t = numpy_vectors_to_torch_tensor(
            arr, shape, torch.float32, torch.device("cpu")
        )
        assert isinstance(t, torch.Tensor)
        assert t.shape == shape
        assert t.device.type == "cpu"
        assert t.dtype == torch.float32

    def test_numpy_vectors_to_torch_tensor_3d(self):
        arr = np.random.rand(2, 4, 3).astype(np.float32)
        shape = (2, 3, 4)
        t = numpy_vectors_to_torch_tensor(
            arr, shape, torch.float32, torch.device("cpu")
        )
        assert isinstance(t, torch.Tensor)
        assert t.shape == shape
        assert t.device.type == "cpu"
        assert t.dtype == torch.float32

    def test_numpy_vectors_to_torch_tensor_invalid_shape(self):
        arr = np.random.rand(
            3,
        )
        with pytest.raises(ValueError):
            numpy_vectors_to_torch_tensor(arr, (3,), torch.float32, torch.device("cpu"))


class TestNpTranformFromTorch:
    def test_np_transform_from_torch(self):
        def dummy_transform(x: np.ndarray) -> np.ndarray:
            return x + 1

        t = torch.arange(6, dtype=torch.float32).reshape(2, 3)
        out = np_transform_from_torch(t, dummy_transform)
        assert torch.allclose(t + 1, out)


class TestMake2DTensor:
    def test_0d_tensor(self):
        t = torch.tensor(7)
        out = make_2d_tensor(t)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (1, 1)
        assert out[0, 0] == 7

    def test_1d_tensor(self):
        t = torch.tensor([1, 2, 3])
        out = make_2d_tensor(t)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (3, 1)
        assert torch.equal(out.squeeze(1), t)

    def test_2d_tensor(self):
        t = torch.arange(6).reshape(2, 3)
        out = make_2d_tensor(t)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (2, 3)
        assert torch.equal(out, t)

    def test_3d_tensor(self):
        t = torch.arange(24).reshape(2, 3, 4)
        out = make_2d_tensor(t)
        # After moveaxis(1, -1): shape (2, 4, 3), then flatten(0, -2): (8, 3)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (8, 3)
        # Check content: first block should match t[0].moveaxis(0, -1).reshape(-1, 3)
        t_moved = t.moveaxis(1, -1)
        t_flat = t_moved.flatten(0, -2)
        assert torch.equal(out, t_flat)

    def test_4d_tensor(self):
        t = torch.arange(2 * 3 * 4 * 5).reshape(2, 3, 4, 5)
        out = make_2d_tensor(t)
        # After moveaxis(1, -1): shape (2, 4, 5, 3), then flatten(0, -2): (40, 3)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (40, 3)
        t_moved = t.moveaxis(1, -1)
        t_flat = t_moved.flatten(0, -2)
        assert torch.equal(out, t_flat)


class DummyIterableDataset(torch.utils.data.IterableDataset):
    def __init__(self, data: list[int]):
        super().__init__()
        self.data = data

    def __iter__(self):
        return iter(self.data)


class TestResetableTorchIterableDataset:
    def test_simple_usage(self):
        dataset = ResetableTorchIterableDataset(DummyIterableDataset(list(range(10))))
        out = list(dataset)
        assert out == list(range(10))

    def test_reset(self):
        dataset = ResetableTorchIterableDataset(DummyIterableDataset(list(range(10))))
        start = next(iter(dataset))
        dataset.reset()
        start_after_reset = next(iter(dataset))
        assert start_after_reset == start



import torch
import pytest
import pixeltable as pxt
from goldener.vectorize import (
    TensorVectorizer,
    Filter2DWithCount,
    FilterLocation,
    GoldVectorizer,
)


class TestTensorsVectorizer:
    def make_tensor(self, shape=(2, 5, 2)):
        return torch.randint(0, 100, shape)

    def test_vectorize_no_y(self):
        x = self.make_tensor()
        v = TensorVectorizer()
        vec = v.vectorize(x)
        assert vec.vectors.shape == (4, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 0, 1, 1]))

    def test_vectorize_with_different_channel_pos(self):
        x = self.make_tensor((2, 2, 5))
        v = TensorVectorizer(channel_pos=2)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (4, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 0, 1, 1]))

    def test_vectorize_with_y(self):
        x = self.make_tensor()
        y = torch.ones(2, 1, 2)
        y[0, 0, 0] = 0
        v = TensorVectorizer()
        vec = v.vectorize(x, y)
        assert vec.vectors.shape == (3, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1, 1]))

    def test_vectorize_with_keep(self):
        x = self.make_tensor()
        keep = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.START, keep=True
        )
        v = TensorVectorizer(keep=keep)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_remove(self):
        x = self.make_tensor()
        remove = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.END, keep=False
        )
        v = TensorVectorizer(remove=remove)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_keep_and_remove(self):
        x = self.make_tensor((2, 5, 3))
        keep = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        remove = Filter2DWithCount(
            filter_count=1, filter_location=FilterLocation.END, keep=False
        )
        v = TensorVectorizer(keep=keep, remove=remove)
        vec = v.vectorize(x)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_with_transform_y(self):
        x = self.make_tensor()
        shape = x.shape
        y = 10 * torch.ones((shape[0], 1, shape[2]))
        y[0, 0, 0] = 3
        y[1, 0, 0] = 3

        def transform_y(y):
            # Only keep rows where y > 5
            return (y > 5).to(torch.int64)

        v = TensorVectorizer(transform_y=transform_y)
        vec = v.vectorize(x, y)
        assert vec.vectors.shape == (2, 5)
        assert torch.equal(vec.batch_indices, torch.tensor([0, 1]))

    def test_vectorize_shape_mismatch(self):
        x = self.make_tensor()
        y = torch.ones(2, 1, 3)
        v = TensorVectorizer()
        with pytest.raises(ValueError):
            v.vectorize(x, y)

    def test_vectorize_2d_input(self):
        x = self.make_tensor((4, 5))
        v = TensorVectorizer()
        with pytest.raises(ValueError):
            v.vectorize(x)

    def test_vectorizer_invalid_keep_type(self):
        with pytest.raises(ValueError):
            TensorVectorizer(keep=Filter2DWithCount())

    def test_vectorizer_invalid_remove_type(self):
        with pytest.raises(ValueError):
            TensorVectorizer(remove=Filter2DWithCount())

    def test_vectorizer_invalid_random_type(self):
        with pytest.raises(ValueError):
            TensorVectorizer(
                random_filter=Filter2DWithCount(filter_location=FilterLocation.START)
            )


class TestFilter2DWithCount:
    def make_tensor(self):
        # 5x3 tensor with unique values for easy row checking
        return torch.arange(15).reshape(5, 3)

    def test_filter_start_keep(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        assert torch.equal(out, x[:2])

    def test_filter_start_remove(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=False
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        assert torch.equal(out, x[2:])

    def test_filter_end_keep(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.END, keep=True
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        assert torch.equal(out, x[-2:])

    def test_filter_end_remove(self):
        x = self.make_tensor()
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.END, keep=False
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        assert torch.equal(out, x[:-2])

    def test_filter_random_keep(self):
        x = self.make_tensor()
        generator = torch.Generator().manual_seed(42)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=True,
            generator=generator,
        )
        out = f.filter(x)
        assert out.shape[0] == 2
        for row in out:
            assert any(torch.equal(row, r) for r in x)

    def test_filter_random_remove(self):
        x = self.make_tensor()
        generator = torch.Generator().manual_seed(42)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=False,
            generator=generator,
        )
        out = f.filter(x)
        assert out.shape[0] == 3
        for row in out:
            assert any(torch.equal(row, r) for r in x)

    def test_filter_tensor_dict(self):
        x = self.make_tensor()
        d = {"a": x.clone(), "b": x.clone() + 100}
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter_tensors(d)
        assert isinstance(out, dict)
        assert set(out.keys()) == {"a", "b"}
        assert torch.equal(out["a"], x[:2])
        assert torch.equal(out["b"], x[:2] + 100)

    def test_filter_random_keep_tensor_dict(self):
        x = self.make_tensor()
        d = {"a": x.clone(), "b": x.clone()}
        generator = torch.Generator().manual_seed(123)
        f = Filter2DWithCount(
            filter_count=2,
            filter_location=FilterLocation.RANDOM,
            keep=True,
            generator=generator,
        )
        out = f.filter_tensors(d)
        assert isinstance(out, dict)
        assert set(out.keys()) == {"a", "b"}
        for tensor in out.values():
            assert tensor.shape[0] == 2
            for row in tensor:
                assert any(torch.equal(row, r) for r in x)

    def test_invalid_filter_count(self):
        with pytest.raises(ValueError):
            Filter2DWithCount(filter_count=0)

    def test_non_2d_input(self):
        x = torch.arange(10)
        f = Filter2DWithCount(filter_count=1)
        with pytest.raises(ValueError):
            f.filter(x)
        d = {"a": torch.arange(10)}
        with pytest.raises(ValueError):
            f.filter_tensors(d)

    def test_inconsistent_batch_size_dict(self):
        x = self.make_tensor()
        d = {"a": x, "b": x[:3]}
        f = Filter2DWithCount(filter_count=2)
        with pytest.raises(ValueError):
            f.filter_tensors(d)

    def test_filter_count_greater_than_rows(self):
        x = self.make_tensor()
        # filter_count > number of rows
        f = Filter2DWithCount(
            filter_count=10, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter(x)
        assert (out == x).all()

    def test_dict_output_keys_and_shapes(self):
        x = self.make_tensor()
        d = {"a": x, "b": x + 1}
        f = Filter2DWithCount(
            filter_count=2, filter_location=FilterLocation.START, keep=True
        )
        out = f.filter_tensors(d)
        assert set(out.keys()) == {"a", "b"}
        assert out["a"].shape == (2, 3)
        assert out["b"].shape == (2, 3)


class DummyDataset:
    def __init__(self, dataset_len: int = 2):
        self.dataset_len = dataset_len

    def __len__(self):
        return self.dataset_len

    def __getitem__(self, idx):
        return {"features": torch.zeros(3, 8, 8), "idx": idx, "label": "dummy"}


class TestGoldVectorizer:
    def test_vectorize_in_table_from_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        pxt.create_dir("unit_test", if_exists="ignore")
        table_path = "unit_test.vectorize_from_dataset"

        dataset = DummyDataset(dataset_len=2)

        gv = GoldVectorizer(
            table_path=table_path,
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            to_keep_schema={"label": pxt.String},
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )

        table = gv.vectorize_in_table(dataset)

        # each sample has 2 vectors (first dim), dataset_len=2 => total 4 rows
        assert table.count() == 8 * 8 * 2
        for row in table.collect():
            assert "vectorized" in row
            assert row["vectorized"].shape == (3,)
            assert row["label"] == "dummy"

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_from_table(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_vectorize"
        desc_path = "unit_test.vectorize_from_table"

        source_rows = [
            {"idx": 0, "features": torch.zeros(4, 3).numpy(), "label": "dummy"},
            {"idx": 1, "features": torch.zeros(4, 3).numpy(), "label": "dummy"},
        ]

        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path, source=source_rows, if_exists="replace_force"
        )

        gv = GoldVectorizer(
            table_path=desc_path,
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            to_keep_schema={"label": pxt.String},
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )

        out_table = gv.vectorize_in_table(src_table)
        assert out_table.count() == 3 * 2
        for row in out_table.collect():
            assert "vectorized" in row
            assert row["vectorized"].shape == (4,)
            assert row["label"] == "dummy"

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_with_target(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_vectorize"
        desc_path = "unit_test.vectorize_from_table"

        source_rows = [
            {
                "idx": 0,
                "features": torch.zeros(4, 3).numpy(),
                "label": "dummy",
                "target": torch.ones(1, 3).numpy(),
            },
            {
                "idx": 1,
                "features": torch.zeros(4, 3).numpy(),
                "label": "dummy",
                "target": torch.ones(1, 3).numpy(),
            },
        ]

        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path, source=source_rows, if_exists="replace_force"
        )

        gv = GoldVectorizer(
            table_path=desc_path,
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            to_keep_schema={"label": pxt.String},
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )

        out_table = gv.vectorize_in_table(src_table)
        assert out_table.count() == 3 * 2
        for row in out_table.collect():
            assert "vectorized" in row
            assert row["vectorized"].shape == (4,)
            assert row["label"] == "dummy"

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_without_idx(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        def collate_fn(batch):
            data = torch.stack([b["features"] for b in batch], dim=0)
            return {"features": data}

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=collate_fn,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )

        table = gv.vectorize_in_table(
            DummyDataset(dataset_len=2),
        )
        assert table.count() == 128
        idx_vector = set()
        idx = set()
        for row in table.collect():
            idx.add(row["idx"])
            idx_vector.add(row["idx_vector"])

        assert idx == {0, 1}
        assert idx_vector == set(range(128))

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_with_non_dict_item(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=lambda x: [d["features"] for d in x],
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )
        with pytest.raises(ValueError, match="Sample must be a dictionary"):
            gv.vectorize_in_table(DummyDataset())

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_with_missing_data_key(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="not_present",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
        )
        with pytest.raises(ValueError, match="Sample is missing expected keys"):
            gv.vectorize_in_table(DummyDataset())

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_with_max_batches(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
            max_batches=2,
        )
        table = gv.vectorize_in_table(
            DummyDataset(dataset_len=3),
        )

        assert table.count() == 128
        for i, row in enumerate(table.collect()):
            assert row["idx_vector"] == i

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        pxt.create_dir("unit_test", if_exists="ignore")
        table_path = "unit_test.vectorize_dataset"

        dataset = DummyDataset(dataset_len=2)

        gv = GoldVectorizer(
            table_path=table_path,
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
            drop_table=True,
        )

        vectorized_dataset = gv.vectorize_in_dataset(dataset)

        count = 0
        for sample in vectorized_dataset:
            assert sample["vectorized"].shape == (3,)
            count += 1

        assert count == 128

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_dataset_from_table(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_vectorize"
        desc_path = "unit_test.vectorize_from_table"

        source_rows = [
            {"idx": 0, "features": torch.zeros(3, 8, 8).numpy()},
            {"idx": 1, "features": torch.zeros(3, 8, 8).numpy()},
        ]
        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path, source=source_rows, if_exists="replace_force"
        )

        gv = GoldVectorizer(
            table_path=desc_path,
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
            drop_table=True,
        )

        vectorized_dataset = gv.vectorize_in_dataset(src_table)

        count = 0
        for sample in vectorized_dataset:
            assert sample["vectorized"].shape == (3,)
            count += 1

        assert count == 128

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_after_restart(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=True,
            max_batches=2,
        )
        dataset = DummyDataset(dataset_len=10)
        vectorized_table = gv.vectorize_in_table(dataset)

        assert vectorized_table.count() == 128
        for i, row in enumerate(vectorized_table.collect()):
            assert row["idx_vector"] == i

        gv.max_batches = None
        vectorized_table = gv.vectorize_in_table(dataset)

        assert vectorized_table.count() == 640
        for i, row in enumerate(vectorized_table.collect()):
            assert row["idx_vector"] == i
            assert row["vectorized"].shape == (3,)

        pxt.drop_dir("unit_test", force=True)

    def test_vectorize_in_table_after_restart_with_restart_disallowed(
        self,
    ):
        pxt.drop_dir("unit_test", force=True)

        gv = GoldVectorizer(
            table_path="unit_test.vectorize",
            vectorizer=TensorVectorizer(),
            collate_fn=None,
            data_key="features",
            vectorized_key="vectorized",
            batch_size=1,
            num_workers=0,
            allow_existing=False,
            max_batches=2,
        )
        dataset = DummyDataset(dataset_len=10)
        gv.vectorize_in_table(dataset)

        with pytest.raises(
            ValueError, match="already exists and allow_existing is set to False"
        ):
            gv.vectorize_in_table(dataset)

        pxt.drop_dir("unit_test", force=True)



from collections import defaultdict
from dataclasses import dataclass
from logging import getLogger
from typing import Callable

import pixeltable as pxt
import torch
from pixeltable.catalog import Table
from torch.utils.data import Dataset, DataLoader

from goldener.describe import GoldDescriptor
from goldener.pxt_utils import (
    get_expr_from_column_name,
    set_value_to_idx_rows,
    GoldPxtTorchDataset,
)
from goldener.select import GoldSelector
from goldener.utils import (
    check_sampling_size,
    check_all_same_type,
    get_sampling_count_from_size,
)
from goldener.vectorize import GoldVectorizer


logger = getLogger(__name__)


@dataclass
class GoldSet:
    """Configuration for a gold set used for splitting.

    Attributes:
        name: Name of the gold set (None if the set is for not selected data).
        size: Size (ratio or actual number) of samples to assign to this set (between 0 and 1). If a ratio, this value
        cannot be one of 0 or 1 (goal is to select a subset of the full dataset). If a number, it must be strictly
        positive and less than the total number of samples in the dataset.
    """

    name: str | None
    size: float | int

    def __post_init__(self) -> None:
        """Validate the GoldSet configuration after initialization.

        Raises:
            ValueError: If ratio is not between 0 and 1 (exclusive).
        """
        if isinstance(self.size, float) and not (0 < self.size < 1):
            raise ValueError("Size as float must be between 0 and 1.")
        elif isinstance(self.size, int) and self.size <= 0:
            raise ValueError("Size as int must be strictly positive.")


def check_sets_validity(sets: list[GoldSet], total: int | None = None) -> None:
    """Validate the sets configuration.

    This private method ensures that the sum of ratios is valid and that set names are unique.

    Args:
        sets: List of GoldSet configurations to validate.
        total: Optional total number of samples (used if sizes are integers).

    Raises:
        ValueError: If the set's sizes are not valid, or if set names are not unique.
    """
    sizes = [s.size for s in sets]
    check_all_same_type(sizes)
    check_sampling_size(sum(sizes), total_size=total)

    set_names = [s.name for s in sets]
    if len(set_names) != len(set(set_names)):
        raise ValueError(f"Set names must be unique, got {set_names}")


class GoldSplitter:
    """Split a dataset into multiple sets based on features.

    The GoldSplitter leverages a GoldDescriptor to extract features from the dataset,
    a GoldVectorizer to vectorize these features, and a GoldSelector to select samples
    for each set based on specified ratios.

    The splitting can operate in a sequential (single-process) mode or a
    distributed mode (not implemented).

    See GoldDescriptor, GoldVectorizer, and GoldSelector for more details on each component.

    Attributes:
        sets: List of GoldSet configurations defining the splits.
        selector: GoldSelector used to select samples for each set. The collate_fn of the selector
        will be set to `pxt_torch_dataset_collate_fn`, and the select_key will be forced to "features"
        to match the descriptor's output column.
        descriptor: Optional GoldDescriptor used to describe the dataset.
        vectorizer: Optional GoldVectorizer used to vectorize the described dataset.
        in_described_table: Whether to return the splitting in the described table or the selected table.
        allow_existing: Whether to allow existing tables in all components.
        drop_table: Whether to drop intermediate tables. Defaults to False.
        max_batches: Optional maximum number of batches to process in both descriptor and selector. Useful for testing on a small subset of the dataset.
    """

    def __init__(
        self,
        sets: list[GoldSet],
        selector: GoldSelector,
        descriptor: GoldDescriptor | None = None,
        vectorizer: GoldVectorizer | None = None,
        in_described_table: bool = False,
        allow_existing: bool = True,
        drop_table: bool = False,
        max_batches: int | None = None,
    ) -> None:
        """Initialize the GoldSplitter.

        Args:
            sets: List of GoldSet configurations defining the splits.
            selector: GoldSelector for selecting samples for each set.
            descriptor: Optional GoldDescriptor for extracting features from the dataset.
            vectorizer: Optional GoldVectorizer for vectorizing described features.
            in_described_table: Whether to return splits in the described table. Defaults to False.
            allow_existing: Whether to allow existing tables. Defaults to True.
            drop_table: Whether to drop intermediate tables. Defaults to False.
            max_batches: Optional maximum number of batches to process.
        """
        if descriptor is None and in_described_table:
            raise ValueError(
                "in_described_table is set to True, but no descriptor is provided."
            )

        if vectorizer is not None:
            if vectorizer.vectorized_key != selector.vectorized_key:
                raise ValueError(
                    f"Vectorizer vectorized_key '{vectorizer.vectorized_key}' does not match "
                    f"selector's vectorized_key '{selector.vectorized_key}'. They must be the same."
                )

            if (
                descriptor is not None
                and descriptor.description_key != vectorizer.data_key
            ):
                raise ValueError(
                    f"Descriptor description_key '{descriptor.description_key}' does not match "
                    f"vectorizer's data_key '{vectorizer.data_key}'. They must be the same."
                )

        elif (
            descriptor is not None
            and descriptor.description_key != selector.vectorized_key
        ):
            raise ValueError(
                f"Descriptor description_key '{descriptor.description_key}' does not match "
                f"selector's vectorized_key '{selector.vectorized_key}'. They must be the same."
            )

        self.sets = sets
        self.descriptor = descriptor
        self.vectorizer = vectorizer
        self.selector = selector
        self.in_described_table = in_described_table
        self.drop_table = drop_table
        self._max_batches = max_batches
        self.allow_existing = allow_existing
        self.max_batches = max_batches

    @property
    def max_batches(self) -> int | None:
        """Get the maximum number of batches to process."""
        return self._max_batches

    @max_batches.setter
    def max_batches(self, value: int | None) -> None:
        """Set the maximum number of batches to process during splitting."""
        self._max_batches = value
        if self.descriptor is not None:
            self.descriptor.max_batches = value
        else:
            if self.vectorizer is not None:
                self.vectorizer.max_batches = value
            else:
                self.selector.max_batches = value

    @property
    def allow_existing(self) -> bool:
        """Get whether existing tables are allowed in all components."""
        return self._allow_existing

    @allow_existing.setter
    def allow_existing(self, value: bool) -> None:
        """Set whether existing tables are allowed in all components."""
        self._allow_existing = value
        if self.descriptor is not None:
            self.descriptor.allow_existing = value
        if self.vectorizer is not None:
            self.vectorizer.allow_existing = value
        self.selector.allow_existing = value

    @property
    def sets(self) -> list[GoldSet]:
        """Get the current sets configuration for the splitter."""
        return self._sets

    @sets.setter
    def sets(self, sets: list[GoldSet]) -> None:
        """Set the sets configuration for the splitter.

        Args:
            sets: New list of GoldSet configurations defining the splits.
        """

        check_sets_validity(sets)
        self._sets = sets

    @staticmethod
    def get_split_indices(
        split_data: Table | Dataset,
        selection_key: str,
        idx_key: str = "idx",
        batch_size: int = 1024,
        num_workers: int = 0,
        collate_fn: Callable | None = None,
    ) -> dict[str, set[int]]:
        """Get the indices of samples for each set from a split table.

        Args:
            split_table: PixelTable Table containing the split information.
            selection_key: Column name in the table indicating the set assignment.
            idx_key: Column name in the table indicating the sample indices.
            batch_size: Batch size to use when processing the input as dataset.
            num_workers: Number of workers to use when processing the input as dataset.
            collate_fn: Collate function to use when processing the input as dataset. It should
                return a dictionary with at least the keys specified by `idx_key` and `selection_key`.

        Returns:
            A dictionary mapping each set name to a set of sample indices.
        """

        set_indices: dict[str, set[int]] = defaultdict(set)

        if isinstance(split_data, Dataset):
            dataloader = DataLoader(
                split_data,
                batch_size=batch_size,
                num_workers=num_workers,
                collate_fn=collate_fn,
            )

            for batch in dataloader:
                batch_indices = batch[idx_key]
                batch_selections = batch[selection_key]
                for idx_value, set_name in zip(batch_indices, batch_selections):
                    if set_name is None:
                        continue
                    set_indices[set_name].add(
                        int(idx_value.item())
                        if isinstance(idx_value, torch.Tensor)
                        else idx_value
                    )
        else:
            selection_col = get_expr_from_column_name(split_data, selection_key)
            for row in split_data.select(selection_col).distinct().collect():
                set_name = row[selection_key]
                if set_name is None:
                    continue

                set_indices[set_name] = GoldSelector.get_selected_sample_indices(
                    split_data,
                    selection_key=selection_key,
                    value=set_name,
                    idx_key=idx_key,
                )

        return set_indices

    def split_in_dataset(
        self,
        to_split: Dataset | Table,
    ) -> GoldPxtTorchDataset:
        """Split the dataset into multiple sets in a PixelTable table.

        The dataset is first described using the gold descriptor (extracts features), and then samples are selected
        for each set based on the specified ratios after vectorization.

        This method is idempotent (i.e. failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will not duplicate or recompute the splitting decisions
        already present in the PixelTable table.

        Args:
            to_split: Dataset or Table to be split. If a Dataset is provided, each item should be a
            dictionary with at least the key specified by descriptor `data_key` after applying the collate_fn.
            If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
            it should contain both 'idx' and `data_key` column for the descriptor.

        Returns:
            A GoldPxtTorchDataset dataset containing at least the set assignation in the `selection_key` column. Then
            the other columns are either from the described table (if `in_described_table` is True)
            or from the selected table (if `in_described_table` is False).
        """
        split_table = self.split_in_table(to_split)

        split_dataset = GoldPxtTorchDataset(split_table, keep_cache=True)

        self._drop_tables(drop_all=True)

        return split_dataset

    def split_in_table(self, to_split: Dataset | Table) -> Table:
        """Split the dataset into multiple sets in a PixelTable table.

        The dataset is first described using the gold descriptor (extracts features), and then samples are selected
        for each set based on the specified ratios after vectorization.

        This method is idempotent (i.e. failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will not duplicate or recompute the splitting decisions
        already present in the PixelTable table.

        See describe_in_table, vectorize_in_table, and select_in_table for more details on each step.

        Args:
            to_split: Dataset or Table to be split. If a Dataset is provided, each item should be a
            dictionary with at least the key specified by descriptor `data_key` after applying the collate_fn.
            If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
            it should contain both 'idx' and `data_key` column for the descriptor.

        Returns:
            A PixelTable Table containing at least the set assignation in the `selection_key` column. Then
            the other columns are either from the described table (if `in_described_table` is True)
            or from the selected table (if `in_described_table` is False).

        Raises:
            ValueError: If any set results in zero samples due to its ratio, if class_key is not found,
            or if class stratification results in zero samples for any class in a set.
        """
        description = (
            self.descriptor.describe_in_table(to_split)
            if self.descriptor is not None
            else to_split
        )

        vectorized = (
            self.vectorizer.vectorize_in_table(description)
            if self.vectorizer is not None
            else description
        )

        if isinstance(vectorized, Table):
            sample_count = vectorized.select(vectorized.idx).distinct().count()
        elif hasattr(vectorized, "__len__"):
            sample_count = len(vectorized)
        else:
            sample_count = None

        # select data for all sets except the last one
        for idx_set, gold_set in enumerate(self._sets):
            logger.info(
                f"Selecting samples for set '{gold_set.name}' with size {gold_set.size}."
            )
            set_count = get_sampling_count_from_size(
                sampling_size=gold_set.size, total_size=sample_count
            )

            selected_table = self.selector.select_in_table(
                vectorized, set_count, value=gold_set.name
            )

        split_table = selected_table
        if self.in_described_table:
            if not isinstance(description, Table):
                raise ValueError(
                    "in_described_table is set to True, but description is not a PixelTable Table."
                )
            for set_cfg in self._sets:
                set_name = set_cfg.name
                set_indices = self.selector.get_selected_sample_indices(
                    selected_table,
                    selection_key=self.selector.selection_key,
                    value=set_name,
                )
                if self.selector.selection_key not in description.columns():
                    description.add_column(
                        if_exists="error",
                        **{self.selector.selection_key: pxt.String},
                    )
                set_value_to_idx_rows(
                    description,
                    col_expr=get_expr_from_column_name(
                        description, self.selector.selection_key
                    ),
                    idx_expr=description.idx,
                    indices=set_indices,
                    value=set_name,
                )
            split_table = description

        self._drop_tables(drop_all=False)

        return split_table

    def _drop_tables(self, drop_all: bool = True) -> None:
        """Drop all intermediate tables created during the splitting process.

        Args:
            drop_all: Whether to drop all tables including the selected table.
            The selected table is either the selector's table or the descriptor's table
            depending on the `in_described_table` flag.
        """
        if self.drop_table:
            table_names = [
                self.selector.table_path,
                *([self.descriptor.table_path] if self.descriptor is not None else []),
                *([self.vectorizer.table_path] if self.vectorizer is not None else []),
            ]

            to_not_drop = (
                self.descriptor.table_path
                if self.in_described_table and self.descriptor is not None
                else self.selector.table_path
            )

            for table in table_names:
                if table != to_not_drop or drop_all:
                    pxt.drop_table(table, if_not_exists="ignore")



from typing import Callable, Any

import numpy as np
import torch
from torch.utils.data import IterableDataset, Dataset


def make_2d_tensor(x: torch.Tensor) -> torch.Tensor:
    """Convert a torch tensor to a 2D tensor.

    If the input tensor is 0D, it is converted to a (1, 1) tensor.
    If the input tensor is 1D, it is converted to a (N, 1) tensor.
    If the input tensor is 2D, it is kept as (N, M) tensor.
    If the input tensor is 3D or higher, the second dimension is moved to the last
    dimension, and the first dimensions are flattened to form a 2D tensor.

    Args:
        x: Input tensor convert to 2D tensor.

    Returns:
        A 2D tensor of shape (N, M)
    """
    initial_ndim = x.ndim
    if initial_ndim == 0:
        x = x.unsqueeze(0).unsqueeze(0)
    elif initial_ndim == 1:
        x = x.unsqueeze(1)

    if x.ndim == 2:
        return x

    x = x.moveaxis(1, -1)
    return x.reshape(-1, x.shape[-1])


def torch_tensor_to_numpy_vectors(x: torch.Tensor) -> np.ndarray:
    """Convert a torch tensor to a numpy array of vectors.

    If the input tensor is 0D, it is converted to a (1, 1) array.
    If the input tensor is 1D, it is converted to a (N, 1) array.
    If the input tensor is 2D, it is kept as (N, M) array.
    If the input tensor is 3D or higher, the second dimension is moved to the last
    dimension, and the first dimensions are flattened to form a 2D array of vectors.

    Args:
        x: Input tensor convert to numpy vectors.

    Returns:
        A 2D numpy array of shape (N, M)
    """
    x = make_2d_tensor(x)
    return x.cpu().numpy()


def numpy_vectors_to_torch_tensor(
    x: np.ndarray, shape: tuple[int, ...], dtype: torch.dtype, device: torch.device
) -> torch.Tensor:
    """Convert a numpy array to a torch tensor.

    If the input array is 0D, it is converted to a (1, 1) tensor.
    If the input array is 1D, it is converted to a (N, 1) tensor.
    If the input array is 2D, it is kept as (N, M) tensor.
    If the input array is 3D or higher, the last dimension is moved to the second
    dimension, and the first dimensions are reshaped to form the desired shape.

    Args:
        x: Input numpy array to convert to torch tensor.
        shape: Desired shape of the output tensor.
        dtype: Desired data type of the output tensor.
        device: Desired device of the output tensor.

    Returns:
        A torch tensor of the desired shape, dtype, and device.

    Raises:
        ValueError: If the desired shape is less than 2D.
    """
    if len(shape) < 2:
        raise ValueError("Shape must be at least 2D")

    if x.ndim <= 2:
        return (
            torch.from_numpy(x)
            .to(
                device=device,
                dtype=dtype,
            )
            .reshape(shape)
        )

    x_torch = torch.from_numpy(x)
    x_torch = x_torch.moveaxis(-1, 1)
    x_torch = x_torch.reshape(shape)

    return x_torch.to(
        device=device,
        dtype=dtype,
    )


def np_transform_from_torch(
    x: torch.Tensor,
    transform_np: Callable[[np.ndarray], np.ndarray],
) -> torch.Tensor:
    """Apply a numpy transformation to a torch tensor.

    Args:
        x: Input tensor to transform.
        transform_np: A callable that takes a numpy array and returns a transformed
            numpy array.

    Returns:
        A transformed torch tensor with the same dtype and device as the input tensor.
    """
    x_np = torch_tensor_to_numpy_vectors(x)
    transformed = transform_np(x_np)
    return numpy_vectors_to_torch_tensor(
        transformed, shape=transformed.shape, dtype=x.dtype, device=x.device
    )


class ResetableTorchIterableDataset(torch.utils.data.IterableDataset):
    """A resetable iterable dataset that can be re-initialized.

    It is useful for accessing the first element (to set up a table for instance)
    of the dataset before looping over the full dataset.

    Attributes:
        data_iterable: The underlying iterable dataset.
        _data_iterator: The current iterator over the dataset.
    """

    def __init__(self, data_iterable: torch.utils.data.IterableDataset):
        """Initialize the resetable iterable dataset.

        Args:
            data_iterable: The underlying iterable dataset to wrap.
        """
        super().__init__()
        self.data_iterable = data_iterable
        self._data_iterator = iter(self.data_iterable)

    def __iter__(self):
        """Return the iterator object."""
        if self._data_iterator is None:
            self._data_iterator = iter(self.data_iterable)
        return self

    def __next__(self):
        """Return the next item from the iterator."""
        try:
            return next(self._data_iterator)
        except StopIteration:
            self._data_iterator = None
            raise

    def reset(self):
        """Reset the iterator to the beginning of the dataset."""
        self._data_iterator = iter(self.data_iterable)


def get_dataset_sample_dict(
    dataset: Dataset,
    collate_fn: Callable | None = None,
    expected: list[str] | None = None,
    excluded: list[str] | None = None,
) -> dict[str, Any]:
    """Get a sample from a dataset as a dictionary and validate some of its keys.

    Args:
        dataset: The dataset to get a sample from.
        collate_fn: An optional collate function to apply to the sample.
        expected: An optional list of keys that must be present in the sample.
        excluded: An optional list of keys that must not be present in the sample.

    Returns:
        A sample from the dataset as a dictionary.
    """

    if isinstance(dataset, IterableDataset):
        dataset = ResetableTorchIterableDataset(dataset)

    sample = (
        next(dataset)
        if isinstance(dataset, ResetableTorchIterableDataset)
        else dataset[0]
    )
    if isinstance(dataset, ResetableTorchIterableDataset):
        dataset.reset()

    if collate_fn is not None:
        sample = collate_fn([sample])

    if not isinstance(sample, dict):
        raise ValueError("Sample must be a dictionary after applying the collate_fn.")

    if expected is not None:
        not_present_keys = [key for key in expected if key not in sample]
        if len(not_present_keys) > 0:
            raise ValueError(f"Sample is missing expected keys: {not_present_keys}")

    if excluded is not None:
        present_rejected_keys = [key for key in excluded if key in sample]
        if len(present_rejected_keys) > 0:
            raise ValueError(f"Sample contains rejected keys: {present_rejected_keys}")

    return sample



import math
from typing import Iterable, Any

import torch


def check_x_and_y_shapes(x_shape: tuple[int, ...], y_shape: tuple[int, ...]) -> None:
    """Check compatibility of shapes of x and y tensors.

    Args:
        x_shape: Shape of the input tensor x.
        y_shape: Shape of the target tensor y.

    Raises:
        ValueError: If the shapes are incompatible. See the conditions below for details.
    Conditions:
        - If x is 1D, x and y must have the same shape.
        - If x is 2D or more, y must have a channel dimension of 1 (i.e., y_shape[1] == 1).
        - If y has a batch size greater than 1 (i.e., y_shape[0] > 1), x and y must have the same batch size (i.e., x_shape[0] == y_shape[0]).
        - If x has more than 2 dimensions, the additional (after channel) dimensions of x and y must match.
    """

    if len(x_shape) == 1:
        if x_shape != y_shape:
            raise ValueError("x and y must have the same shape when x is 1D")
    else:
        if y_shape[1] != 1:
            raise ValueError(
                "x and y must have the same channel dimension when x is 2D"
            )

        batch_y = y_shape[0]
        if batch_y > 1 and x_shape[0] != batch_y:
            raise ValueError(
                "x and y must have the same batch size when y batch size > 1"
            )

        if len(x_shape) > 2 and x_shape[2:] != y_shape[2:]:
            raise ValueError(
                "x and y must have compatible shapes when x is more than 2D"
            )


def get_ratio_list_sum(ratios: list[float]) -> float:
    """Get the sum of a list of ratios and validate it (total between 0 and 1).

    Args:
        ratios: A list of float ratios.

    Returns:
        The sum of the ratios.

    Raises:
        ValueError: If the sum of the ratios is not greater than 0 and at most 1.
    """
    ratio_sum = sum(ratios)
    if not all(ratio >= 0 for ratio in ratios):
        raise ValueError("Ratios must be non-negative.")

    if not (0 < ratio_sum <= 1.0):
        raise ValueError("Sum of ratios must be 1.0.")

    return ratio_sum


def get_ratios_for_counts(counts: Iterable[int]) -> list[float]:
    """Get ratios for a list of counts.

    Args:
        counts: An iterable of integer counts.

    Returns:
        A list of float ratios corresponding to the input counts.
    """
    total = sum(counts)
    return [count / total for count in counts]


def filter_batch_from_indices(
    batch: dict[str, Any],
    to_remove: set[int],
    index_key: str = "idx",
) -> dict[str, Any]:
    """Filter a batch dictionary to only include items at specified indices.

    Args:
        batch: A dictionary representing a batch of data (each key corresponds to stacked information).
        to_remove: A set of indices to filter out from the batch.
        index_key: The key in the batch dictionary that contains the indices.

    Returns:
        A filtered batch dictionary without the specified indices.
    """
    keep_in_batch = [
        idx_position
        for idx_position, idx_value in enumerate(batch[index_key])
        if (idx_value.item() if isinstance(idx_value, torch.Tensor) else idx_value)
        not in to_remove
    ]
    if not keep_in_batch:
        return {}  # all samples already described

    def filter_batched_values(
        batched_value: list | torch.Tensor,
    ) -> list | torch.Tensor:
        """Inner function to remove already described samples from the batch."""
        filtered = [
            value
            for idx_value, value in enumerate(batched_value)
            if idx_value in keep_in_batch
        ]
        if isinstance(batched_value, torch.Tensor):
            return torch.stack(filtered, dim=0)
        else:
            return filtered

    return {
        key: filter_batched_values(batched_value)
        for key, batched_value in batch.items()
    }


def get_size_and_sampling_count_per_chunk(
    total_size: int, sampling_size: int, max_chunk_size: int
) -> tuple[list[int], list[int]]:
    """Get sizes and sampling sizes per chunk.

    Args:
        total_size: Total size of the data to sample from.
        sampling_size: Total number of samples to draw.
        max_chunk_size: Maximum size of each chunk.

    Returns:
        A tuple of two lists:
            - A list of integers representing the size of each chunk.
            - A list of integers representing the number of samples to draw from each chunk.

    Raises:
        ValueError: If sampling_size is greater than or equal to total_size.
    """
    if sampling_size >= total_size:
        raise ValueError("sampling_size must be less than or equal to total_size")

    if max_chunk_size >= total_size:
        return [total_size], [sampling_size]  # single chunk

    chunk_count = math.ceil(total_size / max_chunk_size)
    chunk_size = total_size // chunk_count
    chunk_sampling_size = sampling_size // chunk_count

    chunk_sizes = [chunk_size] * (chunk_count - 1)
    chunk_sizes.append(total_size - sum(chunk_sizes))
    chunk_sampling_sizes = [chunk_sampling_size] * (chunk_count - 1)
    chunk_sampling_sizes.append(sampling_size - sum(chunk_sampling_sizes))

    return chunk_sizes, chunk_sampling_sizes


def check_sampling_size(
    sampling_size: int | float, total_size: int | None = None
) -> None:
    """Check the validity of the sampling size.

    Args:
        sampling_size: The sampling size to check (can be int or float).
        total_size: The total size of the data (Optional).

    Raises:
        ValueError: If the sampling size is invalid based on its type and total size.
        If sampling_size is a float, it must be in the range (0.0, 1.0].
        If sampling_size is an int, it must be in the range (0, total_size].
    """
    if isinstance(sampling_size, float) and not (0 < sampling_size <= 1.0):
        raise ValueError(
            "Sampling size as float must be greater than 0.0 and at most 1.0"
        )

    if (
        isinstance(sampling_size, int)
        and total_size is not None
        and not (0 < sampling_size <= total_size)
    ):
        raise ValueError(
            "Sampling size as int must be greater than 0 and less or equal than the total number of samples"
        )


def check_all_same_type(iterable: Iterable[Any]) -> None:
    """Check if all elements in an iterable are of the same type.

    Args:
        iterable: An iterable containing elements to check.

    Raises:
        TypeError: If any element in the iterable is not of the expected type.
    """
    first_type = type(next(iter(iterable)))
    different = []
    for item in iterable:
        if not isinstance(item, first_type):
            different.append(type(item))

    if different:
        raise TypeError(
            f"All elements must be of the same type {first_type}. Found different types: {different}"
        )


def get_sampling_count_from_size(
    sampling_size: int | float, total_size: int | None = None
) -> int:
    """Get the sampling count from the sampling size.

    Args:
        sampling_size: The sampling size (can be int or float). If int, it represents the exact number of samples
        and it is required to be less than total_size. If float, it represents the fraction of total size and it
        is required to be in the range (0.0, 1.0].
        total_size: The total size of the data (Optional, required if sampling_size is float).

    Returns:
        The calculated sampling count as an integer.

    Raises:
        ValueError: If the sampling size is invalid based on its type and total size.
    """
    if isinstance(sampling_size, int):
        if sampling_size <= 0:
            raise ValueError("Sampling size as int must be greater than 0.")

        if total_size is not None and not (sampling_size < total_size):
            raise ValueError(
                "Sampling size as int must be less than the total number of samples."
            )

        return sampling_size

    if not (0 < sampling_size <= 1.0):
        raise ValueError(
            "Sampling size as float must be greater than 0.0 and at most 1.0."
        )

    if total_size is None:
        raise ValueError("Total size must be provided when sampling size is a float.")

    return math.ceil(sampling_size * total_size)



import torch

from umap import UMAP
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.random_projection import GaussianRandomProjection

from goldener.torch_utils import torch_tensor_to_numpy_vectors, np_transform_from_torch


class GoldReducer:
    """Dimensionality reduction using UMAP, PCA, TSNE, or GaussianRandomProjection.

    Attributes:
        reducer: An instance of UMAP, PCA, TSNE, or GaussianRandomProjection.
    """

    def __init__(self, reducer: UMAP | PCA | TSNE | GaussianRandomProjection):
        """Initialize the GoldReducer.

        Args:
            reducer: An instance of UMAP, PCA, TSNE, or GaussianRandomProjection for dimensionality reduction.
        """
        self.reducer = reducer

    def fit(self, x: torch.Tensor) -> None:
        """Fit the dimensionality reduction model to the data.

        Args:
            x: Input tensor to fit the model on.
        """
        x_np = torch_tensor_to_numpy_vectors(x)
        self.reducer.fit(x_np)

    def fit_transform(self, x: torch.Tensor) -> torch.Tensor:
        """Fit the dimensionality reduction model to the data and transform it.

        Args:
            x: Input tensor to fit and transform.

        Returns:
            Transformed tensor with reduced dimensionality.
        """
        return np_transform_from_torch(x, self.reducer.fit_transform)

    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """Transform the data using the fitted dimensionality reduction model.

        Args:
            x: Input tensor to transform.

        Returns:
            Transformed tensor with reduced dimensionality.
        """
        return np_transform_from_torch(x, self.reducer.transform)



import torch
import pytest
from goldener.utils import (
    check_x_and_y_shapes,
    get_size_and_sampling_count_per_chunk,
    get_ratio_list_sum,
    get_ratios_for_counts,
    filter_batch_from_indices,
    check_sampling_size,
    check_all_same_type,
    get_sampling_count_from_size,
)


class TestCheckXAndYShapes:
    def test_1d_shapes_match(self):
        x = torch.zeros(5)
        y = torch.zeros(5)
        check_x_and_y_shapes(x.shape, y.shape)

    def test_1d_shapes_mismatch(self):
        x = torch.zeros(5)
        y = torch.zeros(6)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_match_channel1(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(3, 1)
        check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_mismatch_channel(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(3, 2)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_mismatch_batch(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(4, 1)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_2d_shapes_match_batch_y_one(self):
        x = torch.zeros(3, 4)
        y = torch.zeros(1, 1)
        check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_match(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 1, 4)
        check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_mismatch_channel(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 2, 4)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_mismatch_batch(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(3, 1, 4)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_mismatch_last_dims(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(2, 1, 5)
        with pytest.raises(ValueError):
            check_x_and_y_shapes(x.shape, y.shape)

    def test_3d_shapes_match_batch_y_one(self):
        x = torch.zeros(2, 3, 4)
        y = torch.zeros(1, 1, 4)
        check_x_and_y_shapes(x.shape, y.shape)

    def test_y_is_none(self):
        x = torch.zeros(2, 3, 4)
        y = None

        if y is not None:
            check_x_and_y_shapes(x.shape, y.shape)


class TestGetSizeAndSamplingCountPerChunk:
    def test_single_chunk_when_max_ge_total(self):
        total_size = 100
        sampling_size = 10
        max_chunk_size = 100
        sizes, samplings = get_size_and_sampling_count_per_chunk(
            total_size, sampling_size, max_chunk_size
        )
        assert sizes == [100]
        assert samplings == [10]

    def test_multiple_chunks_even_split(self):
        total_size = 100
        sampling_size = 10
        max_chunk_size = 30
        sizes, samplings = get_size_and_sampling_count_per_chunk(
            total_size, sampling_size, max_chunk_size
        )
        assert sizes == [25, 25, 25, 25]
        assert samplings == [2, 2, 2, 4]

    def test_sampling_greater_than_total_raises(self):
        with pytest.raises(ValueError):
            get_size_and_sampling_count_per_chunk(10, 11, 5)

    def test_edge_case_small_numbers(self):
        total_size = 5
        sampling_size = 2
        max_chunk_size = 2
        sizes, samplings = get_size_and_sampling_count_per_chunk(
            total_size, sampling_size, max_chunk_size
        )
        assert sizes == [1, 1, 3]
        assert samplings == [0, 0, 2]


class TestCheckSamplingSizes:
    def test_valid_integer_sampling(self):
        check_size = 5
        total_size = 10
        check_sampling_size(check_size, total_size)

    def test_invalid_integer_sampling_raises(self):
        check_size = 15
        total_size = 10
        with pytest.raises(
            ValueError,
            match="Sampling size as int must be greater than 0 and less or equal than the total number of samples",
        ):
            check_sampling_size(check_size, total_size)

    def test_valid_float_sampling(self):
        check_size = 0.5
        check_sampling_size(check_size)

    def test_invalid_float_sampling_raises(self):
        check_size = 1.5
        with pytest.raises(
            ValueError,
            match="Sampling size as float must be greater than 0.0 and at most 1.0",
        ):
            check_sampling_size(check_size)


class TestCheckAllSameType:
    def test_all_same_type(self):
        check_all_same_type([1, 2, 3])

    def test_not_all_same_type(self):
        items = [1, "2", 3]
        with pytest.raises(TypeError, match="All elements must be of the same type"):
            check_all_same_type(items)


class TestGetSamplingCountFromSize:
    def test_integer_sampling_size(self):
        sampling_size = 5
        total_size = 10
        count = get_sampling_count_from_size(
            sampling_size,
            total_size,
        )
        assert count == 5

    def test_float_sampling_size(self):
        sampling_size = 0.5
        total_size = 10
        count = get_sampling_count_from_size(
            sampling_size,
            total_size,
        )
        assert count == 5

    def test_invalid_integer_sampling_raises(self):
        sampling_size = 0
        total_size = 10
        with pytest.raises(
            ValueError, match="Sampling size as int must be greater than 0"
        ):
            get_sampling_count_from_size(
                sampling_size,
                total_size,
            )

        with pytest.raises(
            ValueError,
            match="Sampling size as int must be less than the total number of ",
        ):
            get_sampling_count_from_size(
                10,
                total_size,
            )

    def test_invalid_float_sampling_raises(self):
        sampling_size = 1.0001
        total_size = 10
        with pytest.raises(
            ValueError,
            match="Sampling size as float must be greater than 0.0 and at most 1.0",
        ):
            get_sampling_count_from_size(
                sampling_size,
                total_size,
            )

        with pytest.raises(
            ValueError,
            match="Total size must be provided when sampling size is a float",
        ):
            get_sampling_count_from_size(
                0.5,
                None,
            )


class TestGetRatioListSum:
    def test_valid_ratios_sum_to_one(self):
        ratios = [0.3, 0.3, 0.4]
        result = get_ratio_list_sum(ratios)
        assert result == 1.0

    def test_valid_ratios_sum_less_than_one(self):
        ratios = [0.2, 0.3, 0.4]
        result = get_ratio_list_sum(ratios)
        assert result == 0.9

    def test_invalid_ratios_sum_zero(self):
        ratios = [0.0]
        with pytest.raises(ValueError):
            get_ratio_list_sum(ratios)

    def test_invalid_ratios_negative(self):
        ratios = [-0.1, 0.5]
        with pytest.raises(ValueError, match="Ratios must be non-negative."):
            get_ratio_list_sum(ratios)

    def test_invalid_ratios_sum_greater_than_one(self):
        ratios = [0.6, 0.6]
        with pytest.raises(ValueError, match="Sum of ratios must be 1.0."):
            get_ratio_list_sum(ratios)


class TestGetRatiosForCounts:
    def test_simple_counts(self):
        counts = [10, 20, 30]
        ratios = get_ratios_for_counts(counts)
        assert ratios == [10 / 60, 20 / 60, 30 / 60]


class TestFilterBatchFromIndices:
    def test_filter_some_indices(self):
        batch = {
            "idx": [0, 1, 2, 3, 4],
            "data": ["a", "b", "c", "d", "e"],
        }
        to_remove = {1, 3}
        result = filter_batch_from_indices(batch, to_remove)
        assert result["idx"] == [0, 2, 4]
        assert result["data"] == ["a", "c", "e"]

    def test_filter_no_indices(self):
        batch = {
            "idx": [0, 1, 2],
            "data": ["a", "b", "c"],
        }
        to_remove = set()
        result = filter_batch_from_indices(batch, to_remove)
        assert result["idx"] == [0, 1, 2]
        assert result["data"] == ["a", "b", "c"]

    def test_filter_all_indices(self):
        batch = {
            "idx": [0, 1, 2],
            "data": ["a", "b", "c"],
        }
        to_remove = {0, 1, 2}
        result = filter_batch_from_indices(batch, to_remove)
        assert result == {}

    def test_filter_with_torch_tensors(self):
        batch = {
            "idx": torch.tensor([0, 1, 2, 3]),
            "features": torch.tensor([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]),
        }
        to_remove = {1, 3}
        result = filter_batch_from_indices(batch, to_remove)
        assert torch.equal(result["idx"], torch.tensor([0, 2]))
        assert torch.equal(result["features"], torch.tensor([[1.0, 2.0], [5.0, 6.0]]))

    def test_filter_with_mixed_types(self):
        batch = {
            "idx": torch.tensor([0, 1, 2]),
            "data": ["a", "b", "c"],
            "values": [10, 20, 30],
        }
        to_remove = {1}
        result = filter_batch_from_indices(batch, to_remove)
        assert torch.equal(result["idx"], torch.tensor([0, 2]))
        assert result["data"] == ["a", "c"]
        assert result["values"] == [10, 30]

    def test_filter_with_torch_tensor_indices(self):
        batch = {
            "idx": torch.tensor([10, 20, 30, 40]),
            "data": ["a", "b", "c", "d"],
        }
        to_remove = {20, 40}
        result = filter_batch_from_indices(batch, to_remove)
        assert torch.equal(result["idx"], torch.tensor([10, 30]))
        assert result["data"] == ["a", "c"]

    def test_filter_empty_batch(self):
        batch = {
            "idx": [],
            "data": [],
        }
        to_remove = {0, 1}
        result = filter_batch_from_indices(batch, to_remove)
        assert result == {}



from time import sleep

import numpy as np
import pytest

import torch
from coreax.kernels import LinearKernel
from pixeltable import Error
from sklearn.decomposition import PCA

import pixeltable as pxt
from torch.utils.data import Dataset

from goldener.pxt_utils import GoldPxtTorchDataset
from goldener.reduce import GoldReducer
from goldener.select import GoldSelector
from goldener.select import (
    GoldGreedyClosestPointSelection,
    GoldGreedyKernelPoints,
)


class DummyDataset(Dataset):
    def __init__(self, samples):
        self._samples = samples

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, idx):
        return (
            self._samples[idx].copy()
            if isinstance(self._samples[idx], dict)
            else self._samples[idx]
        )


class TestGoldSelector:
    def test_selection_table_creation_from_table(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_input"
        desc_path = "unit_test.test_select_from_table"

        source_rows = [
            {
                "idx": 0,
                "vectorized": torch.zeros(1, 5).numpy(),
                "label": "dummy",
                "idx_vector": 0,
            },
            {
                "idx_vector": 1,
                "vectorized": torch.zeros(1, 5).numpy(),
                "label": "dummy",
                "idx": 0,
            },
        ]

        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path, source=source_rows, if_exists="replace_force"
        )

        selector = GoldSelector(table_path=desc_path, allow_existing=False)

        pxt_table = selector._selection_table_from_table(
            src_table, old_selection_table=None
        )

        assert set(pxt_table.columns()) == {
            selector.selection_key,
            "idx",
            "idx_vector",
            "chunked",
        }
        row_indices = [
            row["idx_vector"]
            for row in pxt_table.select(pxt_table.idx_vector).collect()
        ]
        assert set(row_indices) == {0, 1}

        pxt.drop_dir("unit_test", force=True)

    def test_selection_table_from_table_when_missing_vectorized(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_invalid"
        pxt.create_dir("unit_test", if_exists="ignore")

        src_table = pxt.create_table(
            src_path,
            source=[{"idx": 0, "notvec": [1, 2, 3]}],
            if_exists="replace_force",
        )

        selector = GoldSelector(table_path="unit_test.test_select", allow_existing=True)

        with pytest.raises(ValueError, match="does not contain the required column"):
            selector._selection_table_from_table(
                select_from=src_table, old_selection_table=None
            )

        pxt.drop_dir("unit_test", force=True)

    def test_selection_table_from_table_when_invalid_old(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.src_table_invalid"
        pxt.create_dir("unit_test", if_exists="ignore")

        src_table = pxt.create_table(
            src_path,
            source=[
                {
                    "idx": 0,
                }
            ],
            if_exists="replace_force",
        )

        selector = GoldSelector(table_path=src_path, allow_existing=True)

        with pytest.raises(ValueError, match="The table is missing required"):
            selector._selection_table_from_table(
                select_from=src_table, old_selection_table=src_table
            )

        pxt.drop_dir("unit_test", force=True)

    def test_selection_table_from_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_initialize"

        sample = {
            "vectorized": torch.rand(1, 5),
            "idx": 0,
        }
        dataset = DummyDataset([sample, sample])

        selector = GoldSelector(table_path=table_path, allow_existing=False)

        pxt_table = selector._selection_table_from_dataset(
            dataset, old_selection_table=None
        )

        assert set(pxt_table.columns()) == {
            selector.selection_key,
            selector.vectorized_key,
            "idx",
            "idx_vector",
            "chunked",
        }
        row_indices = [
            row["idx_vector"]
            for row in pxt_table.select(pxt_table.idx_vector).collect()
        ]
        assert set(row_indices) == {0, 1}

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_from_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=None
        )

        selection_table = selector.select_in_table(
            dataset, select_size=10, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 10
        )

        # validate running with already filled work
        selector.select_in_table(dataset, select_size=10, value="train")

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_wrong_size(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=None
        )

        with pytest.raises(
            ValueError, match="When select_size is a float, it must be in the range"
        ):
            selector.select_in_table(dataset, select_size=1.0, value="train")

        with pytest.raises(ValueError, match="select_size must be a positive integer"):
            selector.select_in_table(dataset, select_size=0, value="train")

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_from_dataset_with_ratio(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=None
        )

        selection_table = selector.select_in_table(
            dataset, select_size=0.1, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 10
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_from_dataset_with_small_ratio(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx_sample": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=None
        )

        selection_table = selector.select_in_table(
            dataset, select_size=0.0001, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 1
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_from_dataset_with_class(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [
                {"vectorized": torch.rand(5), "idx": idx, "label": str(idx % 2)}
                for idx in range(100)
            ]
        )

        selector = GoldSelector(
            table_path=table_path,
            allow_existing=True,
            class_key="label",
            batch_size=10,
            max_batches=None,
        )

        selection_table = selector.select_in_table(
            dataset,
            select_size=10,
            value="train",
        )

        assert selection_table.count() == 100
        assert (
            len(
                selector.get_selected_sample_indices(
                    selection_table,
                    "train",
                    selector.selection_key,
                    class_key=selector.class_key,
                    class_value="0",
                )
            )
            == 5
        )
        assert (
            len(
                selector.get_selected_sample_indices(
                    selection_table,
                    "train",
                    selector.selection_key,
                    class_key=selector.class_key,
                    class_value="1",
                )
            )
            == 5
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_chunk(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_chunk"

        dataset = DummyDataset(
            [
                {
                    "vectorized": torch.rand(
                        5,
                    ),
                    "idx": idx,
                }
                for idx in range(100)
            ]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, chunk=26, batch_size=10
        )

        selection_table = selector.select_in_table(
            dataset, select_size=10, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 10
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_reducer(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_reducer"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path,
            allow_existing=True,
            reducer=GoldReducer(PCA(n_components=3)),
            batch_size=10,
        )

        selection_table = selector.select_in_table(
            dataset, select_size=10, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 10
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_max_batches(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_max_batches"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=2
        )

        selection_table = selector.select_in_table(
            dataset, select_size=10, value="train"
        )

        assert selection_table.count() == 20
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 10
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_restart(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_max_batches"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=True, batch_size=10, max_batches=2
        )

        selector.select_in_table(dataset, select_size=10, value="train")

        selector.max_batches = None
        selection_table = selector.select_in_table(
            dataset, select_size=20, value="train"
        )

        assert selection_table.count() == 100
        assert (
            selection_table.where(
                selection_table[selector.selection_key] == "train"
            ).count()
            == 20
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_restart_disallowed(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_max_batches"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=False, batch_size=10, max_batches=2
        )

        selector.select_in_table(dataset, select_size=10, value="train")

        selector.max_batches = None

        # calling select_in_table when a table exists and allow_existing is False should raise
        with pytest.raises(
            ValueError, match="already exists and allow_existing is set to"
        ):
            selector.select_in_table(dataset, select_size=20, value="train")

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_with_not_enough_sample(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_not_enough"
        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=False, batch_size=10, max_batches=2
        )

        # calling select_in_table when a table exists and allow_existing is False should raise
        with pytest.raises(
            ValueError, match="Cannot select more unique data points than available"
        ):
            selector.select_in_table(dataset, select_size=21, value="train")

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_table_from_table(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.test_select_in_table"

        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path,
            source=[
                {
                    "vectorized": torch.rand(5).numpy().astype(np.float32),
                    "idx": idx % 10,
                    "idx_vector": idx,
                }
                for idx in range(100)
            ],
            if_exists="replace_force",
            primary_key="idx_vector",
        )

        selector = GoldSelector(
            table_path="unit_test.test_select",
            allow_existing=True,
            batch_size=10,
            max_batches=2,
        )

        selector.select_in_table(src_table, select_size=3, value="train")

        selector.max_batches = None
        selection_table = selector.select_in_table(
            src_table, select_size=6, value="train"
        )

        assert selection_table.count() == 100
        assert (
            len(
                selector.get_selected_sample_indices(
                    selection_table, "train", selector.selection_key
                )
            )
            == 6
        )

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_dataset_from_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path, allow_existing=False, batch_size=10, max_batches=2
        )

        dataset = selector.select_in_dataset(dataset, select_count=3, value="train")

        assert isinstance(dataset, GoldPxtTorchDataset)

        selected_count = 0
        total_count = 0
        already_selected = set()
        for item in dataset:
            total_count += 1
            if item["selected"] == "train" and item["idx"] not in already_selected:
                selected_count += 1
                already_selected.add(item["idx"])

        assert total_count == 20
        assert selected_count == 3

        dataset.keep_cache = False

        pxt.drop_dir("unit_test", force=True)

    def test_select_in_dataset_with_drop_table(self):
        pxt.drop_dir("unit_test", force=True)

        table_path = "unit_test.test_select_from_dataset"

        dataset = DummyDataset(
            [{"vectorized": torch.rand(5), "idx": idx} for idx in range(100)]
        )

        selector = GoldSelector(
            table_path=table_path,
            allow_existing=False,
            batch_size=10,
            max_batches=2,
            drop_table=True,
        )

        dataset = selector.select_in_dataset(dataset, select_count=3, value="train")

        sleep(1)
        with pytest.raises(
            Error, match="Path 'unit_test.test_select_from_dataset' does not exist"
        ):
            pxt.get_table(table_path, if_not_exists="error")

        dataset.keep_cache = False

        pxt.drop_dir("unit_test", force=True)

    def test_get_selected_sample_indices(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.test_select"

        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path,
            source=[
                {
                    "vectorized": torch.rand(5).numpy().astype(np.float32),
                    "idx": idx,
                    "selected": "train" if idx < 50 else None,
                    "label": "value" if idx < 25 else "other",
                }
                for idx in range(100)
            ],
            if_exists="replace_force",
        )

        sample_indices = GoldSelector.get_selected_sample_indices(
            table=src_table,
            value="train",
            selection_key="selected",
        )

        assert sample_indices == set(range(50))

        sample_indices = GoldSelector.get_selected_sample_indices(
            table=src_table,
            value=None,
            selection_key="selected",
        )
        assert sample_indices == set(range(50, 100, 1))

        sample_indices = GoldSelector.get_selected_sample_indices(
            table=src_table,
            value="train",
            selection_key="selected",
            class_key="label",
            class_value="value",
        )
        assert sample_indices == set(range(25))

        sample_indices = GoldSelector.get_selected_sample_indices(
            table=src_table,
            value="train",
            selection_key="selected",
            class_key="label",
            class_value="other",
        )
        assert sample_indices == set(range(25, 50, 1))

        with pytest.raises(
            ValueError, match="class_key and class_value must be set together"
        ):
            GoldSelector.get_selected_sample_indices(
                table=src_table,
                value="train",
                selection_key="selected",
                class_key="label",
                class_value=None,
            )

        with pytest.raises(
            ValueError, match="class_key and class_value must be set together"
        ):
            GoldSelector.get_selected_sample_indices(
                table=src_table,
                value="train",
                selection_key="selected",
                class_key=None,
                class_value="other",
            )

        pxt.drop_dir("unit_test", force=True)


class TestGoldGreedyClosestPointSelection:
    def test_basic_selection_properties(self) -> None:
        x = torch.tensor([[0.0], [1.0], [10.0], [10.1], [1.1]], dtype=torch.float32)

        tool = GoldGreedyClosestPointSelection(device="cpu")
        indices = tool.select(x, k=2)

        assert indices == [1, 2]

    def test_select_all_points(self) -> None:
        x = torch.tensor(
            [[0.0], [1.0], [2.0], [3.0]],
            dtype=torch.float32,
        )

        tool = GoldGreedyClosestPointSelection(device="cpu")
        indices = tool.select(x, k=x.size(0))

        assert indices == [0, 1, 2, 3]

    def test_selection_with_multidimensional_features(self) -> None:
        x = torch.tensor(
            [[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]],
            dtype=torch.float32,
        )

        tool = GoldGreedyClosestPointSelection(device="cpu")
        indices = tool.select(x, k=2)

        assert indices == [0, 1]

    def test_with_k_greater_than_size(self) -> None:
        x = torch.tensor([[0.0], [1.0]], dtype=torch.float32)

        tool = GoldGreedyClosestPointSelection(device="cpu")
        with pytest.raises(
            ValueError, match="k cannot be greater than the number of data points in x"
        ):
            tool.select(x, k=5)


class TestGoldGreedyKernelPoints:
    def test_basic_selection_with_linear_kernel(self) -> None:
        tool = GoldGreedyKernelPoints(
            feature_kernel=LinearKernel(output_scale=1, constant=0)
        )

        x = torch.arange(12, dtype=torch.float32).view(4, 3)
        k = 2

        indices = tool.select(x, k)

        assert len(indices) == k
        assert len(set(indices)) == k

    def test_select_all_points_with_linear_kernel(self) -> None:
        tool = GoldGreedyKernelPoints(
            feature_kernel=LinearKernel(output_scale=1, constant=0)
        )

        x = torch.arange(12, dtype=torch.float32).view(4, 3)
        k = 4

        indices = tool.select(x, k)

        assert len(indices) == k
        assert len(set(indices)) == k

    def test_with_k_greater_than_size(self) -> None:
        x = torch.tensor([[0.0], [1.0]], dtype=torch.float32)

        tool = GoldGreedyKernelPoints(
            feature_kernel=LinearKernel(output_scale=1, constant=0)
        )
        with pytest.raises(ValueError, match="must be less than"):
            tool.select(x, k=5)



from abc import abstractmethod
from typing_extensions import assert_never
from typing import Dict, List, Callable, Any

from dataclasses import dataclass
from enum import Enum

import torch


class GoldFeatureExtractor:
    """Abstract base class for feature extraction from models.

    This class defines the interface for feature extractors that can extract and optionally
    fuse features from models. Implementations should provide specific mechanisms for
    extracting features from different types of models (e.g., PyTorch, multimodal).
    """

    @abstractmethod
    def extract(self, *args: Any, **kwargs: Any) -> dict[str, torch.Tensor]:
        """Extract features from the model for the given input data.

        Returns: Dictionary mapping layer names to their extracted feature tensors.
        """

    @abstractmethod
    def extract_and_fuse(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        """Extract and fuse features from the model for the given input data.

        Returns: Fused feature tensor.
        """


class FeatureFusionStrategy(Enum):
    """Strategies to fuse features from multiple layers.

    CONCAT: Concatenate features along the channel dimension.
    ADD: Element-wise addition of features.
    AVERAGE: Element-wise average of features.
    MAX: Element-wise maximum of features.
    NO_FUSION: No fusion, return features as is (only valid for single layer).
    """

    CONCAT = "concat"
    ADD = "add"
    AVERAGE = "average"
    MAX = "max"


@dataclass
class TorchGoldFeatureExtractorConfig:
    """Configuration for the TorchFeatureExtractor.

    Attributes:
        model: The PyTorch model from which to extract features.
        layers: List of layer names or a dictionary mapping group names to lists of layer names.
            If None, the last layer of the model is used.
        layer_fusion: Strategy to fuse features from multiple layers within the same group.
        group_fusion: Strategy to fuse features from different groups.
    """

    model: torch.nn.Module
    layers: list[str] | dict[str, list[str]] | None = None
    layer_fusion: FeatureFusionStrategy = FeatureFusionStrategy.CONCAT
    group_fusion: FeatureFusionStrategy = FeatureFusionStrategy.CONCAT


class GoldFeatureFusion:
    """Feature fusion from multiple layers and groups.

    Attributes:
        layer_fusion: Strategy to fuse features from multiple layers within the same group.
        group_fusion: Strategy to fuse features from different groups.
    """

    def __init__(
        self,
        layer_fusion: FeatureFusionStrategy = FeatureFusionStrategy.CONCAT,
        group_fusion: FeatureFusionStrategy = FeatureFusionStrategy.CONCAT,
    ) -> None:
        """Initialize the GoldFeatureFusion.

        Args:
            layer_fusion: Strategy to fuse features from multiple layers within the same group.
                Defaults to CONCAT.
            group_fusion: Strategy to fuse features from different groups. Defaults to CONCAT.
        """
        self.layer_fusion = layer_fusion
        self.group_fusion = group_fusion

    @staticmethod
    def fuse_tensors(
        tensors: List[torch.Tensor],
        strategy: FeatureFusionStrategy,
    ) -> torch.Tensor:
        """Fuse a list of tensors.

        The tensors are at least expected to have the shape (B, C), until (B, C, D, H, W) if sizes after the
        channel dimension differ. In this case, all tensors are interpolated to the largest size.

        Args:
            tensors: List of tensors to be fused.
            strategy: Strategy to fuse the tensors.

        Returns: Fused tensors.

        Raises:
            ValueError: If the tensors have a different number of dimensions.
        """
        ndims = set(f.ndim for f in tensors)
        if len(ndims) != 1:
            raise ValueError("All features must have the same number of dimensions.")

        ndim = list(ndims)[0]
        if ndim > 2:
            max_size = tuple(
                max(sizes) for sizes in zip(*(f.shape[2:] for f in tensors))
            )
            # Interpolate all tensors to the largest size
            mode = "linear" if ndim == 3 else ("bilinear" if ndim == 4 else "trilinear")
            tensors = [
                (
                    torch.nn.functional.interpolate(
                        feature,
                        size=max_size,
                        mode=mode,
                    )
                    if feature.shape[2:] != max_size
                    else feature
                )
                for feature in tensors
            ]

        if strategy is FeatureFusionStrategy.CONCAT:
            return torch.cat(tensors, dim=1)
        elif strategy is FeatureFusionStrategy.ADD:
            return torch.stack(tensors, dim=0).sum(dim=0)
        elif strategy is FeatureFusionStrategy.AVERAGE:
            return torch.stack(tensors, dim=0).mean(dim=0)
        elif strategy is FeatureFusionStrategy.MAX:
            return torch.stack(tensors, dim=0).max(dim=0).values
        else:
            assert_never(strategy)

    def fuse_features(
        self,
        x: dict[str, torch.Tensor],
        layers: list[str] | dict[str, list[str]],
    ) -> torch.Tensor:
        """Fuse features from multiple layers and groups.

        Args:
            x: Dictionary mapping layer names to feature tensors.
            layers: List of layer names or a dictionary mapping group names to lists of layer names.

        Returns: Fused feature tensor.
        """
        # list of layers are fused by layer_fusion strategy
        if isinstance(layers, list):
            return self.fuse_tensors([x[name] for name in layers], self.layer_fusion)

        # groups of layers are fused by layer_fusion strategy,
        # then all groups are fused by group_fusion strategy
        fused_groups = []
        for group, layer_names in layers.items():
            if len(layer_names) == 1:
                fused_groups.append(x[layer_names[0]])
            else:
                fused_groups.append(
                    self.fuse_tensors(
                        [x[name] for name in layer_names], self.layer_fusion
                    )
                )
        return self.fuse_tensors(fused_groups, self.group_fusion)


class TorchGoldFeatureExtractor(GoldFeatureExtractor):
    """Feature extractor for PyTorch models.

    Once initialized, the extractor registers forward hooks on the specified layers of the model.
    When the model processes input data, the hooks capture the outputs of these layers.
    The extracted features can then be fused according to the specified strategies.

    The model and layers cannot be changed after initialization. The feature fusion can be changed.

    Attributes:
        _model: The PyTorch model from which to extract features.
        fusion: FeatureFusion instance to handle feature fusion.
        _layers: List of layer names or a dictionary mapping group names to lists of layer names.
        _hooks: Dictionary mapping layer names to their corresponding forward hook handles.
        _features: Dictionary to store extracted features.
    """

    def __init__(
        self,
        config: TorchGoldFeatureExtractorConfig,
    ) -> None:
        """Initialize the TorchGoldFeatureExtractor.

        Args:
            config: Configuration object containing the model, layers, and fusion strategies.
        """
        self._model = config.model
        self.fusion = GoldFeatureFusion(
            layer_fusion=config.layer_fusion,
            group_fusion=config.group_fusion,
        )

        self._layers: list[str] | dict[str, list[str]]
        self._hooks: dict[str, torch.utils.hooks.RemovableHandle]
        self._features: dict[str, torch.Tensor]

        self._register_layers(config.layers)

    @property
    def model(self) -> torch.nn.Module:
        """The PyTorch model from which to extract features."""
        return self._model

    @property
    def layers(self) -> list[str] | dict[str, list[str]]:
        """The layers from which to extract features.

        It also indicates the grouping of layers for fusion.
        """
        return self._layers

    def extract(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        """Extract features from the model for the given input data.

        Args:
            x: Input data tensor to be processed by the model.

        Returns: Dictionary mapping layer names to their extracted feature tensors.
        """
        self._features = {}
        self._model(x)
        return self._features

    def extract_and_fuse(self, x: torch.Tensor) -> torch.Tensor:
        """Extract and fuse features from the model for the given input data.

        Args:
            x: Input data tensor to be processed by the model.

        Returns: Fused feature tensor.
        """
        features = self.extract(x)
        return self.fusion.fuse_features(features, self._layers)

    def __del__(self):
        """Remove all registered hooks when the extractor is deleted."""
        for handle in self._hooks.values():
            handle.remove()

    def _register_layers(self, layers: List[str] | Dict[str, List[str]] | None) -> None:
        """Register forward hooks on the specified layers of the model.

        Args:
            layers: List of layer names or a dictionary mapping group names to lists of layer names.
                If None, the last layer of the model is used.
        """

        named_modules = list(self._model.named_modules())
        if layers is None:
            layer_names = [named_modules[-1][0]]  # last layer
            layers = layer_names
        elif isinstance(layers, dict):
            layer_names = [name for names in layers.values() for name in names]
        elif isinstance(layers, list):
            layer_names = layers
        else:
            assert_never(layers)

        self._layers = layers

        self._features = {}

        def _get_hook(
            name: str,
        ) -> Callable[[torch.nn.Module, torch.Tensor, torch.Tensor], None]:
            def hook(
                module: torch.nn.Module,
                input: torch.Tensor,
                output: torch.Tensor,
            ) -> None:
                self._features[name] = output.detach()

            return hook

        self._hooks = {
            name: module.register_forward_hook(_get_hook(name))
            for name, module in named_modules
            if name in layer_names
        }

        not_found = set(layer_names).difference(set(self._hooks.keys()))
        if not_found:
            raise ValueError(f"Layers not found in the model: {not_found}")


class MultiModalTorchGoldFeatureExtractor(GoldFeatureExtractor):
    """Feature extractor for multimodal data using PyTorch.

    Each modality has its own TorchFeatureExtractor defined by its own configuration.
    This allows for processing different types of input data (e.g., images, text, audio)
    with different models and then fusing their features.

    Attributes:
        extractors: Dictionary mapping modality names to their TorchGoldFeatureExtractor instances.
        strategy: Strategy for fusing features from different modalities.
    """

    def __init__(
        self,
        configs: Dict[str, TorchGoldFeatureExtractorConfig],
        strategy: FeatureFusionStrategy = FeatureFusionStrategy.CONCAT,
    ) -> None:
        """Initialize the multimodal feature extractor.

        Args:
            configs: Dictionary mapping modality names to their TorchGoldFeatureExtractorConfig.
            strategy: Strategy to use for fusing features from different modalities. Defaults to CONCAT.
        """
        self.extractors = {
            modality: TorchGoldFeatureExtractor(config)
            for modality, config in configs.items()
        }
        self.strategy = strategy

    def extract_and_fuse(self, x: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Extract and fuse features from multimodal input data.

        Args:
            x: Dictionary mapping modality names to their input tensors.

        Returns:
            Fused feature tensor combining all modalities.
        """
        return GoldFeatureFusion.fuse_tensors(
            [
                extractor.extract_and_fuse(x[modality])
                for modality, extractor in self.extractors.items()
            ],
            self.strategy,
        )

    def extract(self, x: Dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Extract features from multimodal input data without fusing.

        Args:
            x: Dictionary mapping modality names to their input tensors.

        Returns:
            Dictionary mapping "{modality}.{layer}" to their extracted feature tensors.
        """
        per_modality = {
            modality: extractor.extract(x[modality])
            for modality, extractor in self.extractors.items()
        }

        return {
            f"{modality}.{layer}": feature
            for modality, features in per_modality.items()
            for layer, feature in features.items()
        }



import pytest

import torch

from goldener.extract import (
    GoldFeatureFusion,
    FeatureFusionStrategy,
    TorchGoldFeatureExtractor,
    TorchGoldFeatureExtractorConfig,
    MultiModalTorchGoldFeatureExtractor,
)


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(3, 4, 3, padding=1)
        self.conv2 = torch.nn.Conv2d(4, 8, 3, padding=1)
        self.relu = torch.nn.ReLU()

    def forward(self, x):
        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        return x


def make_tensor(shape, fill_value=None):
    if fill_value is not None:
        return torch.full(shape, fill_value)
    return torch.randn(shape)


shapes_2d_3d_4d = [
    (2, 4),
    (2, 4, 5),
    (2, 4, 5, 5),
    (2, 4, 5, 5, 5),
]


class TestFeatureFusion:
    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_concat(self, shape):
        t1 = make_tensor(shape)
        t2 = make_tensor(shape)
        fused = GoldFeatureFusion.fuse_tensors([t1, t2], FeatureFusionStrategy.CONCAT)
        assert fused.shape[1] == shape[1] * 2
        assert fused.shape[:2] == (shape[0], shape[1] * 2)[:2]

    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_add(self, shape):
        t1 = make_tensor(shape, 1.0)
        t2 = make_tensor(shape, 1.0)
        fused = GoldFeatureFusion.fuse_tensors([t1, t2], FeatureFusionStrategy.ADD)
        assert torch.allclose(fused, torch.full_like(fused, 2.0))

    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_average(self, shape):
        t1 = make_tensor(shape, 1.0)
        t2 = make_tensor(shape, 3.0)
        fused = GoldFeatureFusion.fuse_tensors([t1, t2], FeatureFusionStrategy.AVERAGE)
        assert torch.allclose(fused, torch.full_like(fused, 2.0))

    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_with_different_shapes(self, shape):
        # Only test for 3D and 4D shapes (spatial dims)
        if len(shape) < 3:
            return
        t1 = make_tensor(shape)
        smaller_shape = (shape[0], shape[1]) + tuple(max(1, s // 2) for s in shape[2:])
        t2 = make_tensor(smaller_shape)
        fused = GoldFeatureFusion.fuse_tensors([t1, t2], FeatureFusionStrategy.ADD)
        assert fused.shape == shape

    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_fuse_features(self, shape):
        t1 = make_tensor(shape)
        t2 = make_tensor(shape)
        features = {"mod1": t1, "mod2": t2}
        fusion = GoldFeatureFusion(
            layer_fusion=FeatureFusionStrategy.ADD,
            group_fusion=FeatureFusionStrategy.ADD,
        )
        fused = fusion.fuse_features(features, ["mod1", "mod2"])
        assert fused.shape == shape

    @pytest.mark.parametrize("shape", shapes_2d_3d_4d)
    def test_fuse_features_with_groups(self, shape):
        t1 = make_tensor(shape)
        t2 = make_tensor(shape)
        t3 = make_tensor(shape)
        features = {"layer1": t1, "layer2": t2, "layer3": t3}
        fusion = GoldFeatureFusion(
            layer_fusion=FeatureFusionStrategy.ADD,
            group_fusion=FeatureFusionStrategy.CONCAT,
        )
        fused = fusion.fuse_features(
            features, {"m1": ["layer1", "layer2"], "m2": ["layer3"]}
        )
        # Should concatenate along channel dim
        assert fused.shape[1] == shape[1] * 2
        assert fused.shape[0] == shape[0]
        if len(shape) > 2:
            assert fused.shape[2:] == shape[2:]


class TestTorchFeatureExtractor:
    def test_extract(self):
        model = DummyModel()
        layers = ["conv1", "conv2"]
        config = TorchGoldFeatureExtractorConfig(model=model, layers=layers)
        extractor = TorchGoldFeatureExtractor(config)
        data = torch.randn(2, 3, 8, 8)
        features = extractor.extract(data)
        assert len(features) == len(layers)
        assert features["conv1"].shape == (2, 4, 8, 8)
        assert features["conv2"].shape == (2, 8, 8, 8)

    def test_extract_and_fuse(self):
        model = DummyModel()
        layers = ["conv1", "conv2"]
        config = TorchGoldFeatureExtractorConfig(
            model=model,
            layers=layers,
            layer_fusion=FeatureFusionStrategy.CONCAT,
        )
        extractor = TorchGoldFeatureExtractor(config)
        data = torch.randn(2, 3, 8, 8)
        fused = extractor.extract_and_fuse(data)
        # Should add features from conv1 and conv2
        assert fused.shape == (2, 12, 8, 8)

    def test_invalid_layer(self):
        model = DummyModel()
        config = TorchGoldFeatureExtractorConfig(model=model, layers=["invalid_layer"])
        with pytest.raises(ValueError):
            TorchGoldFeatureExtractor(config)


class TestMultiModalTorchFeatureExtractor:
    def test_extract(self):
        model1 = DummyModel()
        model2 = DummyModel()
        config1 = TorchGoldFeatureExtractorConfig(model=model1, layers=["conv1"])
        config2 = TorchGoldFeatureExtractorConfig(model=model2, layers=["conv2"])
        extractor = MultiModalTorchGoldFeatureExtractor(
            {"img": config1, "aux": config2}
        )
        data = {
            "img": torch.randn(2, 3, 8, 8),
            "aux": torch.randn(2, 3, 8, 8),
        }
        features = extractor.extract(data)
        assert len(features) == 2
        assert features["img.conv1"].shape == (2, 4, 8, 8)
        assert features["aux.conv2"].shape == (2, 8, 8, 8)

    def test_extract_and_fuse(self):
        model1 = DummyModel()
        model2 = DummyModel()
        config1 = TorchGoldFeatureExtractorConfig(model=model1, layers=["conv1"])
        config2 = TorchGoldFeatureExtractorConfig(model=model2, layers=["conv2"])
        extractor = MultiModalTorchGoldFeatureExtractor(
            {"img": config1, "aux": config2},
            strategy=FeatureFusionStrategy.CONCAT,
        )
        data = {
            "img": torch.randn(2, 3, 8, 8),
            "aux": torch.randn(2, 3, 8, 8),
        }
        fused = extractor.extract_and_fuse(data)
        # Should concatenate features from both modalities
        assert fused.shape[0] == 2
        assert fused.shape[2:] == (8, 8)



from dataclasses import dataclass
from functools import partial
from logging import getLogger

from pixeltable import Error
from torch.utils.data import RandomSampler, Dataset, DataLoader
from tqdm import tqdm
from typing_extensions import assert_never
from typing import Callable, Any

from enum import Enum

import torch
from torch import Generator

import pixeltable as pxt
from pixeltable.catalog import Table

from goldener.pxt_utils import (
    GoldPxtTorchDataset,
    get_valid_table,
    get_sample_row_from_idx,
    pxt_torch_dataset_collate_fn,
    get_expr_from_column_name,
    make_batch_ready_for_table,
    check_pxt_table_has_primary_key,
    get_max_value_in_column,
)
from goldener.torch_utils import make_2d_tensor, get_dataset_sample_dict
from goldener.utils import check_x_and_y_shapes, filter_batch_from_indices

logger = getLogger(__name__)


class FilterLocation(Enum):
    """Enum defining filter location strategies for filtering 2D tensor rows.

    START: Filter from the start of the tensor.
    END: Filter from the end of the tensor.
    RANDOM: Filter randomly from the tensor.
    """

    START = "start"
    END = "end"
    RANDOM = "random"


class Filter2DWithCount:
    """Filter 2D tensor rows based on specified criteria.

    Attributes:
        _filter_location: Location to filter from (start, end, random).
        _filter_count: Number of rows to filter.
        _keep: Whether to keep or remove the filtered rows.
        _random_sampler: Sampler for random filtering.
    """

    def __init__(
        self,
        filter_count: int = 1,
        filter_location: FilterLocation = FilterLocation.RANDOM,
        keep: bool = False,
        generator: Generator | None = None,
    ) -> None:
        """Initialize the Filter2DWithCount.

        Args:
            filter_count: Number of rows to filter.
            filter_location: Location to filter from (start, end, random).
            keep: Whether to keep or remove the filtered rows.
            generator: Random number generator for random filtering.
        """
        if filter_count <= 0:
            raise ValueError("filter_count must be greater than 0")

        self._filter_location = filter_location
        self._filter_count = filter_count
        self._keep = keep

        if self._filter_location == FilterLocation.RANDOM:
            self._random_sampler = partial(
                RandomSampler,
                replacement=False,
                num_samples=self._filter_count,
                generator=generator,
            )

    @property
    def is_random(self) -> bool:
        """Check if the filter is random.

        Returns:
            True if the filter uses random sampling, False otherwise.
        """
        return self._filter_location is FilterLocation.RANDOM

    def filter(
        self,
        x: torch.Tensor,
    ) -> torch.Tensor:
        """Filter a single 2D tensor.

        Args:
            x: Input 2D tensor to filter.

        Returns:
            Filtered tensor.
        """
        return self.filter_tensors({"x": x})["x"]

    def filter_tensors(
        self,
        x: dict[str, torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        """Filter the input tensor or dictionary of tensors.

        Args:
            x: Input 2D tensor or dictionary of 2D tensors to filter. In case of a dictionary,
            all tensors must have the same number of rows and this number must be greater than filter_count.

        Returns:
            Filtered tensor or dictionary of tensors. If the batch size is less than filter_count,
            the input is returned unchanged.

        Raises:
            ValueError: If input tensors do not have the same batch size or are not 2 dimensional.
        """
        first_tensor = next(iter(x.values()))
        batch_size = first_tensor.shape[0]
        for tensor in x.values():
            if tensor.ndim != 2:
                raise ValueError(
                    "All input tensors must be 2D to filter them with Filter2DWithCount."
                )

            if tensor.shape[0] != batch_size:
                raise ValueError(
                    "All input tensors must have the same batch size to filter them with Filter2DWithCount."
                )

        if batch_size < self._filter_count:
            return x

        if self._filter_location is FilterLocation.START:
            return self._start_filter(x)

        elif self._filter_location is FilterLocation.END:
            return self._end_filter(x)

        elif self._filter_location is FilterLocation.RANDOM:
            return self._random_filter(x)

        else:
            assert_never(self._filter_location)

    def _start_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Filter rows from the start of tensors."""
        return (
            {k: v[: self._filter_count] for k, v in x.items()}
            if self._keep
            else {k: v[self._filter_count :] for k, v in x.items()}
        )

    def _end_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Filter rows from the end of tensors."""
        return (
            {k: v[-self._filter_count :] for k, v in x.items()}
            if self._keep
            else {k: v[: -self._filter_count] for k, v in x.items()}
        )

    def _random_filter(self, x: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Filter rows randomly from tensors."""
        first_value = next(iter(x.values()))
        indices = list(self._random_sampler(range(len(first_value))))
        if self._keep:
            return {k: v[indices] for k, v in x.items()}
        else:
            mask = torch.ones(len(first_value), dtype=torch.bool)
            mask[indices] = 0
            return {k: v[mask.bool(), :] for k, v in x.items()}


@dataclass
class Vectorized:
    """Dataclass to hold vectorized tensors and the corresponding batch indices.

    Attributes:
        vectors: 2D tensor of vectorized data.
        batch_indices: 1D tensor containing information about the origin of each vector.
    """

    vectors: torch.Tensor
    batch_indices: torch.Tensor


class TensorVectorizer:
    """Transform input as 2D tensor and filter based on target tensor.

    Attributes:
        keep: Filter2DWithCount instance to keep specific rows in the input `x` of `filter`.
        remove: Filter2DWithCount instance to remove specific rows in the input `x` of `filter`.
        random_filter: Random Filter2DWithCount instance to randomly filter vectors randomly after
        applying `keep`, `remove` and the target `y` on the input `x` of filter.
        transform_y: Optional callable to transform the target tensor before transforming it to 2D.
        channel_pos: position of the channel dimension in the input tensor to vectorize.
    """

    def __init__(
        self,
        keep: Filter2DWithCount | None = None,
        remove: Filter2DWithCount | None = None,
        random_filter: Filter2DWithCount | None = None,
        transform_y: Callable[[torch.Tensor], torch.Tensor] | None = None,
        channel_pos: int = 1,
    ) -> None:
        """Initialize the TensorVectorizer.

        Args:
            keep: Optional filter to keep specific rows in the input.
            remove: Optional filter to remove specific rows from the input.
            random_filter: Optional random filter to apply after keep/remove filters.
            transform_y: Optional transformation to apply to the target tensor.
            channel_pos: Position of the channel dimension in the input tensor. Defaults to 1.

        Raises:
            ValueError: If keep or remove filters are random, or if random_filter is not random.
        """
        self.transform_y = transform_y
        self.channel_pos = channel_pos

        if keep is not None and keep.is_random:
            raise ValueError("The 'keep' filter cannot be random.")
        self.keep = keep

        if remove is not None and remove.is_random:
            raise ValueError("The 'remove' filter cannot be random.")
        self.remove = remove

        if random_filter is not None and not random_filter.is_random:
            raise ValueError("The 'random_filter' must be random.")
        self.random_filter = random_filter

    def vectorize(
        self,
        x: torch.Tensor,
        y: torch.Tensor | None = None,
    ) -> Vectorized:
        """Vectorize input tensor and filter based on target tensor.

        If y is not provided, only filtering and vectorization are performed. If y is provided,
        it is used to filter the vectorized x tensor.

        Args:
            x: Input tensor to vectorize.
            y: Optional target tensor to filter the input tensor.

        Returns:
            Vectorized and filtered input tensor with the information
            about which element of the batch it corresponds to.

        Raises:
            ValueError: If x and y shapes are incompatible. See check_x_and_y_shapes for details.
        """
        if x.ndim < 3:
            raise ValueError("Input tensor x to vectorize must be at least 3D.")

        if y is not None and self.transform_y is not None:
            y = self.transform_y(y)

        if self.channel_pos != 1:
            # Move channel dimension to position 1
            x = x.movedim(self.channel_pos, 1)

        filtered_x = []
        filtered_batch_info = []
        for idx_sample, x_sample in enumerate(x):
            x_sample = x_sample.unsqueeze(0)
            x_sample = make_2d_tensor(x_sample)

            x_sample = self._apply_filter(
                self.keep.filter if self.keep is not None else None, x_sample
            )
            x_sample = self._apply_filter(
                self.remove.filter if self.remove is not None else None, x_sample
            )

            if y is not None:
                y_sample = y[idx_sample].unsqueeze(0)
                x_sample = self._filter_2d_tensors_from_y(x_sample, y_sample)

            filtered_x.append(x_sample)
            filtered_batch_info.append(
                torch.full_like(x_sample[:, 0], idx_sample, dtype=torch.long)
            )

        return Vectorized(
            torch.cat(filtered_x, dim=0), torch.cat(filtered_batch_info, dim=0)
        )

    def _apply_filter(
        self,
        filter: Callable[[torch.Tensor], torch.Tensor] | None,
        x: torch.Tensor,
    ) -> torch.Tensor:
        """Apply a filter function to a tensor if provided.

        Args:
            filter: Optional filter function to apply.
            x: Input tensor.

        Returns:
            Filtered tensor, or original tensor if filter is None.
        """
        if filter is None:
            return x

        return filter(x)

    def _filter_2d_tensors_from_y(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> torch.Tensor:
        """Filter 2D tensor rows based on target tensor.

        When y is not the same batch size as x, it is expanded to match x's batch size.

        Args:
            x: Input 2D tensor to filter. It can be the output of `make_2d_tensor`.
            y: Target tensor to filter the input tensor. It is still in its raw configuration.
            The transform_y callable is applied to y before transforming it to a 2D tensor.

        Returns:
            Filtered input tensor.

        Raises:
            ValueError: If x and y shapes (after 2D transformation) are incompatible.
            See check_x_and_y_shapes for details.
            ValueError: If y tensor after transform contains only zeros.
        """
        y = make_2d_tensor(y)

        y_shape = y.shape
        x_shape = x.shape
        check_x_and_y_shapes(x_shape, y_shape)

        if (y == 0).all():
            raise ValueError(
                "The y tensor after transform must contain at least one "
                "non-zero value to select vectors from x."
            )

        return x[y.bool().squeeze(-1)]


class GoldVectorizer:
    """Extract and flatten vectors from dataset samples and store results in a PixelTable table.

    The GoldVectorizer processes a dataset or PixelTable table to extract and flatten vectors using a
    `TensorVectorizer`. The computed vectors are stored in a local PixelTable table
    (specified by `table_path`) so that the vectorization process is idempotent: calling the
    same operation multiple times will not duplicate or recompute vectors that are already
    present in the table.

    Assuming all the data will not fit in memory, the dataset is processed in batches.
    All torch tensors will be converted to numpy arrays before saving.

    The vectorization can operate in a sequential (single-process) mode or a
    distributed mode (not implemented). The table schema is created/validated automatically and will include
    minimal indexing columns (`idx`, `idx_vector`) required to link vectors back to
    their originating samples.

    Attributes:
        table_path: Path to the PixelTable table where vectorized outputs will be stored locally.
        vectorizer: TensorVectorizer instance for transforming batched inputs into vectors.
        collate_fn: Optional function to collate dataset samples into batches composed of
            dictionaries with at least the key specified by `data_key` returning a PyTorch Tensor.
            If None, the dataset is expected to directly provide such batches.
        data_key: Key in the batch dictionary that contains the data to vectorize. Default is "features".
        target_key: Optional key in the batch dictionary containing the target used to filter vectors. Default is "target".
        vectorized_key: Column name to store the resulting vectors in the PixelTable table. Default is "vectorized".
        to_keep_schema: Optional dictionary defining additional columns to keep from the original dataset/table
            into the vectorized table. The keys are the column names and the values are the PixelTable types.
        min_pxt_insert_size: Minimum number of rows to accumulate before inserting into the PixelTable table. Default is 100.
        batch_size: Batch size used when iterating over the data.
        num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
        allow_existing: If False, an error will be raised when the table already exists. Default is True.
        distribute: Whether to use distributed processing for vectorization and table population. Not implemented yet. Default is False.
        drop_table: Whether to drop the vectorized table after creating the dataset with vectorized outputs. It is only applied
            when using `vectorize_in_dataset`. Default is False.
        max_batches: Optional maximum number of batches to process. Useful for testing on a small subset of the dataset.
    """

    _MINIMAL_SCHEMA: dict[str, type] = {
        "idx": pxt.Required[pxt.Int],
        "idx_vector": pxt.Required[pxt.Int],
    }

    def __init__(
        self,
        table_path: str,
        vectorizer: TensorVectorizer,
        collate_fn: Callable | None = None,
        data_key: str = "features",
        target_key: str = "target",
        vectorized_key: str = "vectorized",
        to_keep_schema: dict[str, type] | None = None,
        min_pxt_insert_size: int = 100,
        batch_size: int = 1,
        num_workers: int = 0,
        allow_existing: bool = True,
        distribute: bool = False,
        drop_table: bool = False,
        max_batches: int | None = None,
    ) -> None:
        """Initialize the GoldVectorizer.

        Args:
            table_path: Path to the PixelTable table for storing vectors.
            vectorizer: TensorVectorizer instance for transforming tensors.
            collate_fn: Optional collate function for preparing batches.
            data_key: Key for data in the batch dictionary. Defaults to "features".
            target_key: Key for target in the batch dictionary. Defaults to "target".
            vectorized_key: Column name for storing vectors. Defaults to "vectorized".
            to_keep_schema: Optional schema for additional columns to preserve.
            min_pxt_insert_size: Minimum number of rows to accumulate before inserting into the PixelTable table. Default is 100.
            batch_size: Batch size used when iterating over the data.
            num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
            allow_existing: Whether to allow using an existing table. Defaults to True.
            distribute: Whether to use distributed processing. Defaults to False.
            drop_table: Whether to drop the table after dataset creation. Defaults to False.
            max_batches: Optional maximum number of batches to process.
        """
        self.table_path = table_path
        self.vectorizer = vectorizer
        self.collate_fn = collate_fn
        self.data_key = data_key
        self.target_key = target_key
        self.vectorized_key = vectorized_key
        self.to_keep_schema = to_keep_schema
        self.min_pxt_insert_size = min_pxt_insert_size
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.allow_existing = allow_existing
        self.distribute = distribute
        self.drop_table = drop_table
        self.max_batches = max_batches

    def vectorize_in_dataset(
        self,
        to_vectorize: Dataset | Table,
    ) -> GoldPxtTorchDataset:
        """Extract and flatten vectors from samples and return results as a GoldPxtTorchDataset.

        The vectorization process extracts and flattens vectors from the provided dataset or table
        using the `TensorVectorizer` instance and stores them in a PixelTable table specified by `table_path`.
        The resulting vectors are stored in the column specified by `vectorized_key`.

        This is a convenience wrapper that runs `vectorize_in_table` to populate
        (or resume populating) the PixelTable table, then wraps the table into a
        `GoldPxtTorchDataset` for downstream consumption. If `drop_table` is True,
        the table will be removed after the dataset is created.

        Args:
            to_vectorize: Dataset or Table to be vectorized. If a Dataset is provided, each item should be a
                dictionary with at least the key specified by `data_key` after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
                it should contain both 'idx' and `data_key` columns.

        Returns:
            A GoldPxtTorchDataset containing at least the vectorized data in the `vectorized_key` key
                and `idx` (index of the sample) and `idx_vector` (index of the vector) keys as well.
        """
        vectorized_table = self.vectorize_in_table(to_vectorize)

        vectorized_dataset = GoldPxtTorchDataset(vectorized_table, keep_cache=True)

        if self.drop_table:
            pxt.drop_table(vectorized_table)

        return vectorized_dataset

    def vectorize_in_table(
        self,
        to_vectorize: Dataset | Table,
    ) -> Table:
        """Extract and flatten vectors from samples and store results in a PixelTable table.

        The vectorization process extracts and flattens vectors from the provided dataset or table
        using the `TensorVectorizer` instance and stores them in a PixelTable table specified by `table_path`.
        The resulting vectors are stored in the column specified by `vectorized_key`.

        This method is idempotent (i.e., failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will not duplicate or recompute the vectorization
        already present in the PixelTable table.

        Args:
            to_vectorize: Dataset or Table to be vectorized. If a Dataset is provided, each item should be a
                dictionary with at least the key specified by `data_key` after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches. If a Table is provided,
                it should contain both 'idx' and `data_key` columns.

        Returns:
            A PixelTable Table containing at least the vectorized data in the `vectorized_key` column
                and `idx` (index of the sample) and `idx_vector` (index of the vector) columns as well.
        """

        # If the computation was already started or already done, we resume from there
        logger.info(f"Loading the existing vectorized table from {self.table_path}")
        try:
            old_vectorized_table = pxt.get_table(
                self.table_path,
                if_not_exists="ignore",
            )
        except Error:
            logger.info(f"No existing vectorized table from {self.table_path}")
            old_vectorized_table = None

        # vectorized table are expected to have a primary key allowing to idempotent updates
        if old_vectorized_table is not None:
            check_pxt_table_has_primary_key(old_vectorized_table, set(["idx_vector"]))

        if not self.allow_existing and old_vectorized_table is not None:
            raise ValueError(
                f"Table at path {self.table_path} already exists and "
                "allow_existing is set to False."
            )

        # get the table and dataset to execute the vectorize pipeline
        to_vectorize_dataset: Dataset | GoldPxtTorchDataset
        if isinstance(to_vectorize, Table):
            vectorized_table = self._vectorized_table_from_table(
                to_vectorize=to_vectorize,
                old_vectorized_table=old_vectorized_table,
            )

            if vectorized_table.count() > 0 and "idx" in to_vectorize.columns():
                to_vectorize_indices = set(
                    [
                        row["idx"]
                        for row in to_vectorize.select(to_vectorize.idx).collect()
                    ]
                )
                already_vectorized = set(
                    [
                        row["idx"]
                        for row in vectorized_table.select(vectorized_table.idx)
                        .distinct()
                        .collect()
                    ]
                )
                if not to_vectorize_indices.difference(already_vectorized):
                    logger.info(
                        f"Vectorized table already fully filled out from {self.table_path}"
                    )
                    return vectorized_table

            to_vectorize_dataset = GoldPxtTorchDataset(to_vectorize)
        else:
            to_vectorize_dataset = to_vectorize
            vectorized_table = self._vectorized_table_from_dataset(
                to_vectorize, old_vectorized_table
            )

        if self.distribute:
            vectorized = self._distributed_vectorize(
                vectorized_table, to_vectorize_dataset
            )
        else:
            vectorized = self._sequential_vectorize(
                vectorized_table, to_vectorize_dataset
            )

        logger.info(
            f"Vectorized table populated {vectorized.count()} rows at {self.table_path}"
        )

        return vectorized

    def _vectorized_table_from_table(
        self, to_vectorize: Table, old_vectorized_table: Table | None
    ) -> Table:
        """Create or validate the vectorized table schema from a PixelTable table.

        This private method sets up the table structure and adds the vectorized column
        with the appropriate array type based on the vectorizer's output shape.

        Args:
            to_vectorize: The source PixelTable table to vectorize.
            old_vectorized_table: Existing vectorized table if resuming, or None.

        Returns:
            The vectorized table with proper schema.
        """
        vectorized_table = self._get_vectorized_valid_table(old_vectorized_table)

        if self.vectorized_key not in vectorized_table.columns():
            logger.info(f"Add the Vectorized column in {self.table_path}")
            sample = get_sample_row_from_idx(
                to_vectorize,
                collate_fn=pxt_torch_dataset_collate_fn,
                expected_keys=[self.data_key],
            )
            vectorized = self.vectorizer.vectorize(
                sample[self.data_key],
                sample.get(self.target_key, None),
            ).vectors[0]
            vectorized_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized.shape, pxt.Float
                    ]
                }
            )

        return vectorized_table

    def _vectorized_table_from_dataset(
        self, to_vectorize: Dataset, old_vectorized_table: Table | None
    ) -> Table:
        """Create or validate the vectorized table schema from a PyTorch Dataset.

        This private method sets up the table structure and adds the vectorized column
        with the appropriate array type based on the vectorizer's output shape.

        Args:
            to_vectorize: The source PyTorch Dataset to vectorize.
            old_vectorized_table: Existing vectorized table if resuming, or None.

        Returns:
            The vectorized table with proper schema.
        """
        vectorized_table = self._get_vectorized_valid_table(old_vectorized_table)

        if self.vectorized_key not in vectorized_table.columns():
            logger.info(f"Add the Vectorized column in {self.table_path}")
            sample = get_dataset_sample_dict(
                to_vectorize,
                collate_fn=self.collate_fn,
                expected=[self.data_key],
                excluded=[self.vectorized_key],
            )

            target = sample.get(self.target_key, None)
            if target is not None and self.collate_fn is not None:
                assert isinstance(target, torch.Tensor)
                target = target.unsqueeze(0)

            vectorized = (
                self.vectorizer.vectorize(
                    sample[self.data_key].unsqueeze(0)
                    if self.collate_fn is None
                    else sample[self.data_key],
                    target,
                )
                .vectors[0]
                .detach()
                .cpu()
                .numpy()
            )
            vectorized_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized.shape, pxt.Float
                    ]
                }
            )

        return vectorized_table

    def _get_vectorized_valid_table(self, old_vectorized_table: Table | None) -> Table:
        minimal_schema = self._MINIMAL_SCHEMA
        if self.to_keep_schema is not None:
            minimal_schema |= self.to_keep_schema

        return get_valid_table(
            table=old_vectorized_table
            if old_vectorized_table is not None
            else self.table_path,
            minimal_schema=minimal_schema,
            primary_key="idx_vector",
        )

    def _distributed_vectorize(
        self,
        vectorized_table: Table,
        to_vectorize_dataset: Dataset,
    ) -> Table:
        """Run distributed vectorization process (not implemented).

        Args:
            vectorized_table: The table to store vectorized outputs.
            to_vectorize_dataset: The dataset to vectorize.

        Returns:
            The populated vectorized table.

        Raises:
            NotImplementedError: Always raised as distributed mode is not yet implemented.
        """
        raise NotImplementedError("Distributed Vectorization is not implemented yet.")

    def _sequential_vectorize(
        self,
        vectorized_table: Table,
        to_vectorize_dataset: Dataset,
    ) -> Table:
        """Run sequential (single-process) vectorization process.

        This private method processes the dataset in batches, applies vectorization using
        the TensorVectorizer, and stores the results in the vectorized table. It is idempotent
        and will skip samples that have already been vectorized.

        Args:
            vectorized_table: The table to store vectorized outputs.
            to_vectorize_dataset: The dataset to vectorize.

        Returns:
            The populated vectorized table.
        """
        assert self.batch_size is not None
        assert self.num_workers is not None

        not_empty = (
            vectorized_table.count() > 0
        )  # allow to filter out already described samples

        vectorized_col = get_expr_from_column_name(
            vectorized_table, self.vectorized_key
        )
        already_vectorized = set(
            [
                row["idx"]
                for row in vectorized_table.where(
                    vectorized_col != None  # noqa: E711
                )
                .select(vectorized_table.idx)
                .collect()
            ]
        )

        dataloader = DataLoader(
            to_vectorize_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
        )

        ready_to_insert: list[dict[str, Any]] = []

        start_idx = (
            0
            if vectorized_table.count() == 0
            else get_max_value_in_column(vectorized_table, vectorized_table.idx_vector)
            + 1
        )

        for batch_idx, batch in tqdm(
            enumerate(dataloader),
            desc="Vectorizing the dataset",
            total=(
                None
                if hasattr(to_vectorize_dataset, "__len__") is False
                else len(dataloader)
            ),
        ):
            # Stop if we've processed enough batches
            if self.max_batches is not None and batch_idx >= self.max_batches:
                break

            # add idx if it is not provided by the dataset
            if "idx" not in batch:
                starts = batch_idx * self.batch_size
                batch["idx"] = [
                    starts + idx for idx in range(len(batch[self.data_key]))
                ]

            # Keep only not yet described samples in the batch
            if not_empty:
                batch = filter_batch_from_indices(
                    batch,
                    already_vectorized,
                )

                if len(batch) == 0:
                    continue  # all samples already described

            already_vectorized.update(
                [
                    idx.item() if isinstance(idx, torch.Tensor) else idx
                    for idx in batch["idx"]
                ]
            )

            # vectorize data
            to_keep_keys = (
                list(self.to_keep_schema.keys())
                if self.to_keep_schema is not None
                else []
            )
            batch = vectorize_and_unwrap_in_batch(
                batch=batch,
                vectorizer=self.vectorizer,
                data_key=self.data_key,
                vectorized_key=self.vectorized_key,
                target_key=self.target_key,
                to_keep=to_keep_keys,
                starts=start_idx,
            )

            start_idx = max(batch["idx_vector"]) + 1

            to_insert_keys = [self.vectorized_key, "idx"] + to_keep_keys
            batch_as_list = make_batch_ready_for_table(
                batch,
                to_insert_keys,
                "idx_vector",
            )
            ready_to_insert.extend(batch_as_list)
            if len(ready_to_insert) >= self.min_pxt_insert_size:
                vectorized_table.batch_update(ready_to_insert, if_not_exists="insert")
                ready_to_insert = []

        if ready_to_insert:
            vectorized_table.batch_update(ready_to_insert, if_not_exists="insert")

        return vectorized_table


def unwrap_vectors_in_batch(
    vectorized: Vectorized,
    vectorized_key: str,
    batch: dict[str, Any],
    starts: int = 0,
    to_keep: list[str] | None = None,
) -> dict[str, Any]:
    """Unwrap vectorized output into individual rows for table insertion.

    This converts the Vectorized dataclass (which contains batched vectors
    and their batch indices) into a dictionary format suitable for inserting into the
    PixelTable table, with one entry per vector.

    Args:
        vectorized: Vectorized output containing vectors and batch indices.
        vectorized_key: Key under which to store the vectors in the output dictionary.
        batch: Original batch dictionary with metadata.
        starts: Starting index for assigning new `idx_vector` values.
        to_keep: Optional list of additional keys to preserve from the batch.

    Returns:
        Dictionary with unwrapped vectors and metadata, ready for table insertion.
    """
    new_batch: dict[str, Any] = {
        "idx": [],
        vectorized_key: [],
        "idx_vector": [],
    }
    if to_keep is not None:
        for key in to_keep:
            new_batch[key] = []

    sample_indices = [
        (idx_value.item() if isinstance(idx_value, torch.Tensor) else idx_value)
        for batch_idx, idx_value in enumerate(batch["idx"])
    ]

    vectorized_idx = vectorized.batch_indices
    for vec_idx, vector in enumerate(vectorized.vectors):
        new_batch[vectorized_key].append(vector)
        new_batch["idx_vector"].append(starts + vec_idx)
        batch_idx = vectorized_idx[vec_idx].item()
        assert isinstance(batch_idx, int)
        new_batch["idx"].append(sample_indices[batch_idx])
        if to_keep is not None:
            for key in to_keep:
                new_batch[key].append(batch[key][batch_idx])

    return new_batch


def vectorize_and_unwrap_in_batch(
    batch: dict[str, Any],
    vectorizer: TensorVectorizer,
    data_key: str,
    vectorized_key: str,
    target_key: str | None,
    to_keep: list[str] | None = None,
    starts: int = 0,
) -> dict[str, Any]:
    """Vectorize a batch and insert the results into a PixelTable table.

    This function vectorizes the data in the provided batch using the specified
    TensorVectorizer and inserts the resulting vectors into the given PixelTable table.

    Args:
        vectorized_table: The PixelTable table to insert vectorized outputs into.
        batch: The batch dictionary containing data to vectorize.
        vectorizer: The TensorVectorizer instance to use for vectorization.
        data_key: Key in the batch dictionary that contains the data to vectorize.
        vectorized_key: Column name to store the resulting vectors in the PixelTable table.
        target_key: Optional key in the batch dictionary containing the target used to filter vectors.
        to_keep: Optional list of additional keys to preserve from the batch.
    """
    target = (
        batch[target_key] if target_key is not None and target_key in batch else None
    )

    vectorized = vectorizer.vectorize(
        batch[data_key],
        target,
    )

    batch = unwrap_vectors_in_batch(
        vectorized=vectorized,
        vectorized_key=vectorized_key,
        batch=batch,
        starts=starts,
        to_keep=to_keep,
    )

    return batch



import torch

from goldener.reduce import GoldReducer
from umap import UMAP
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.random_projection import GaussianRandomProjection


def test_pca():
    data = torch.randn(10, 5, dtype=torch.float32)
    reducer = PCA(n_components=3)
    dr = GoldReducer(reducer)
    # Test fit + transform
    dr.fit(data)
    out = dr.transform(data)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (data.shape[0], 3)
    # Test fit_transform
    out2 = dr.fit_transform(data)
    assert isinstance(out2, torch.Tensor)
    assert out2.shape == (data.shape[0], 3)


def test_tsne():
    data = torch.randn(10, 5, dtype=torch.float32)
    reducer = TSNE(n_components=2, perplexity=3)
    dr = GoldReducer(reducer)
    # TSNE does not have transform, only fit_transform
    out = dr.fit_transform(data)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (data.shape[0], 2)


def test_umap():
    data = torch.randn(10, 5, dtype=torch.float32)
    reducer = UMAP(n_components=2)
    dr = GoldReducer(reducer)
    # Test fit + transform
    dr.fit(data)
    out = dr.transform(data)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (data.shape[0], 2)
    # Test fit_transform
    out2 = dr.fit_transform(data)
    assert isinstance(out2, torch.Tensor)
    assert out2.shape == (data.shape[0], 2)


def test_gaussian_random_projection():
    data = torch.randn(10, 5, dtype=torch.float32)
    reducer = GaussianRandomProjection(n_components=4)
    dr = GoldReducer(reducer)
    # Test fit + transform
    dr.fit(data)
    out = dr.transform(data)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (data.shape[0], 4)
    # Test fit_transform
    out2 = dr.fit_transform(data)
    assert isinstance(out2, torch.Tensor)
    assert out2.shape == (data.shape[0], 4)



from collections import defaultdict
from typing import Literal, Any, Iterator, Sequence, Callable
import shutil

import numpy as np
import pixeltable as pxt
from pixeltable import Query
from pixeltable.catalog import Table
from pixeltable.exprs import Expr
import pixeltable.functions as pxtf

import torch
from pixeltable.type_system import ArrayType
from pixeltable.utils.pytorch import PixeltablePytorchDataset

from goldener.utils import get_ratios_for_counts


def get_expr_from_column_name(table: Table, column_name: str) -> Expr:
    """Get the expression object for a given column name in a PixelTable table.

    Args:
        table: The PixelTable table.
        column_name: The name of the column.

    Returns:
        pxt.Expr: The expression object corresponding to the column name.

    Raises:
        ValueError: If the column name does not exist in the table schema.
    """
    if column_name not in table.columns():
        raise ValueError(f"Column '{column_name}' does not exist in the table schema.")

    return getattr(table, column_name)


def create_pxt_dirs_for_path(table_path: str) -> None:
    """Create necessary PixelTable directories for a given table path.

    Args:
        table_path: The full path of the PixelTable table (e.g., 'dir1.dir2.table_name').
    """
    pxt_path_split = table_path.split(".")
    for pxt_dir_idx in range(len(pxt_path_split) - 1):
        pxt_dir = ".".join(pxt_path_split[0 : pxt_dir_idx + 1])
        pxt.create_dir(pxt_dir, if_exists="ignore")


def create_pxt_table_from_sample(
    table_path: str,
    sample: dict,
    unwrap: bool = False,
    add: dict[str, Any] | None = None,
    if_exists: Literal["error", "replace_force"] = "error",
) -> Table:
    """Create a PixelTable table from a sample dictionary.

    Args:
        table_path: The full path of the PixelTable table to create (e.g., 'dir1.dir2.table_name').
        sample: A dictionary representing a single sample, where keys are column names and values are the corresponding data.
        unwrap: Whether to unwrap (only take first element) tensors/lists in the sample.
        add: Additional key-value pairs to add to the sample before creating the table.
        if_exists: Behavior if the table already exists. Options are 'error' or 'replace_force'.
    """
    create_pxt_dirs_for_path(table_path)

    if unwrap:
        for key, value in sample.items():
            if isinstance(value, torch.Tensor):
                # Only initial tensors will have a batch dimension added by the collate_fn
                # otherwise, they were initially single values (float or int)
                if value.ndim > 1:
                    sample[key] = value.squeeze(0)
                else:
                    sample[key] = value.item()
            else:
                # non tensor values are expected to be lists of single values
                sample[key] = value[0]

    if add is not None:
        for key, value in add.items():
            if key in sample:
                raise ValueError(f"Key '{key}' already exists in the sample.")
            sample[key] = value

    pxt_table = pxt.create_table(
        table_path,
        source=[
            {
                key: (value.numpy() if isinstance(value, torch.Tensor) else value)
                for key, value in sample.items()
            }
        ],
        if_exists=if_exists,
    )

    pxt_table.where(pxt_table.idx == 0).delete()  # remove the initial sample

    return pxt_table


def set_value_to_idx_rows(
    table: Table,
    col_expr: Expr,
    idx_expr: Expr,
    indices: set[int],
    value: int | float | str | None,
) -> None:
    """Set a column to a specific value for rows with given indices in a PixelTable table.

    Args:
        table: The PixelTable table. Must contain an 'idx' column.
        col_expr: The column expression to be set.
        idx_expr: The column expression for the index.
        indices: List of row indices to update.
        value: The value to set the column to.
    """
    table.where(idx_expr.isin(indices)).update({col_expr.display_str(): value})


def pxt_torch_dataset_collate_fn(batch: list[dict[str, Any]]) -> dict[str, Any]:
    """Collate function for torch datasets obtained from a Pixeltable table.

    Args:
        batch: A list of samples, where each sample is a dictionary.

    Returns:
        A single dictionary with collated values. The numpy arrays and integers
        are stacked into torch tensors. All other types are kept as lists.
    """
    value_list_dict = defaultdict(list)
    conversion_dict: dict[str, Callable[[Sequence[Any]], torch.Tensor]] = {}

    def stack_arrays_and_convert_to_torch(arrays: Sequence[np.ndarray]) -> torch.Tensor:
        return torch.from_numpy(np.stack(arrays, axis=0))

    def stack_int_as_torch(ints: Sequence[int]) -> torch.Tensor:
        return torch.tensor(ints, dtype=torch.int64)

    for idx_sample, sample in enumerate(batch):
        for key, value in sample.items():
            value_list_dict[key].append(value)
            if idx_sample == 0:
                if isinstance(value, np.ndarray):
                    conversion_dict[key] = stack_arrays_and_convert_to_torch
                elif isinstance(value, int):
                    conversion_dict[key] = stack_int_as_torch

    return {
        key: conversion_dict[key](value) if key in conversion_dict else value
        for key, value in value_list_dict.items()
    }


class GoldPxtTorchDataset(PixeltablePytorchDataset):
    """A Pixeltable PyTorch Dataset that reshapes array columns to their original shapes.

    This class intends to solve a current issue in Pixeltable dataset where array columns
    are flattened when converted to PyTorch datasets. This class reshapes those columns
    back to their original shapes during iteration.

    Attributes:
        shapes: A dictionary mapping column names of arrays to their original shapes.
        keep_cache: Whether to keep the cached parquet files after the dataset is garbage collected.
        path: Path to directory containing cached parquet files.
        image_format: Image format ('np' or 'pt').
    """

    def __init__(self, source: Table | Query, keep_cache: bool = False) -> None:
        """Initialize the GoldPxtTorchDataset.

        Args:
            source: The PixelTable Table or Query to convert to a PyTorch dataset.
            keep_cache: Whether to keep the cached parquet files after the dataset is garbage collected.
        """

        self.shapes = (
            get_array_column_shapes_from_table(source)
            if isinstance(source, Table)
            else get_array_column_shapes_from_query(source)
        )
        self.keep_cache = keep_cache

        torch_dataset = source.to_pytorch_dataset()
        assert isinstance(torch_dataset, PixeltablePytorchDataset)
        super().__init__(torch_dataset.path, torch_dataset.image_format)

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """Iterate over the dataset, reshaping array columns to their original shapes.

        Yields:
            Dictionary representing a single sample with properly shaped array columns.
        """
        super_iterator = super().__iter__()
        for item in super_iterator:
            yield {
                key: value
                if key not in self.shapes
                else value.reshape(self.shapes[key])
                for key, value in item.items()
            }

    def __del__(self) -> None:
        """Clean up cached parquet files when the dataset is destroyed.

        This method removes the cache directory created by PixeltablePytorchDataset
        to prevent accumulation of temporary files.
        """
        if self.path.exists() and not self.keep_cache:
            shutil.rmtree(self.path, ignore_errors=True)


def get_array_column_shapes_from_table(table: Table) -> dict[str, tuple[int, ...]]:
    """Get the shapes of array columns in a PixelTable table.

    Args:
        table: The PixelTable table.

    Returns:
        A dictionary mapping column names of array columns to their shapes.
    """
    shapes = {}
    for col_name in table.columns():
        col_expr = get_expr_from_column_name(table, col_name)
        col_expr_dict = col_expr.col_type.as_dict()
        if col_expr_dict["_classname"] == "ArrayType":
            shapes[col_name] = tuple(col_expr_dict["shape"])

    return shapes


def get_array_column_shapes_from_query(query: Query) -> dict[str, tuple[int, ...]]:
    """Get the shapes of array columns in a PixelTable Query.

    Args:
        query: The PixelTable Query.

    Returns:
        A dictionary mapping column names of array columns to their shapes.
    """
    shapes = {}

    for col_name, col_type in query.schema.items():
        if isinstance(col_type, ArrayType):
            shape = col_type.shape
            assert shape is not None
            shapes[col_name] = tuple(s for s in shape if s is not None)

    return shapes


def get_distinct_value_and_count_in_column(
    table: Table,
    col_expr: Expr,
) -> dict[Any, int]:
    """Get distinct values and their counts in a specified column of a PixelTable table.

    Args:
        table: The PixelTable table.
        col_expr: The column expression to analyze.

    Returns:
        A dictionary mapping distinct values to their counts in the specified column.
    """
    return {
        distinct_item[col_expr.display_str()]: table.where(
            col_expr == distinct_item[col_expr.display_str()]
        ).count()
        for distinct_item in table.select(col_expr).distinct().collect()
    }


def get_column_distinct_ratios(table: Table, class_expr: Expr) -> dict[str, float]:
    """Get the ratios of distinct values in a specified column of a PixelTable table.

    Args:
        table: The PixelTable table.
        class_expr: The column expression to analyze.

    Returns:
        A dictionary mapping distinct values to their ratios in the specified column.
    """
    value_and_count = get_distinct_value_and_count_in_column(table, class_expr)
    ratios = get_ratios_for_counts(value_and_count.values())

    return {
        class_label: class_ratio
        for class_label, class_ratio in zip(value_and_count.keys(), ratios)
    }


def get_sample_row_from_idx(
    table: Table,
    idx: int = 0,
    collate_fn: Callable | None = None,
    expected_keys: list[str] | None = None,
) -> dict[str, Any]:
    """Get a sample row from a PixelTable table.

    Args:
        table: The PixelTable table.
        idx: The index of the row to retrieve.
        collate_fn: An optional collate function to apply to the sample.
        expected_keys: An optional list of expected keys in the sample.

    Returns:
        A dictionary representing a sample row from the table.
    """
    sample_list = list(table.where(table.idx == idx).collect())

    if not sample_list:
        raise ValueError(f"No sample found at the index {idx}.")

    if len(sample_list) > 1:
        raise ValueError("Multiple samples found at the specified index.")

    sample = sample_list[0]
    if expected_keys is not None:
        not_present_keys = [key for key in expected_keys if key not in sample]
        if len(not_present_keys) > 0:
            raise ValueError(f"Sample is missing expected keys: {not_present_keys}")

    if collate_fn:
        sample = collate_fn([sample])

    return sample


def get_valid_table(
    table: Table | str,
    minimal_schema: dict[str, type],
    primary_key: str | None = None,
) -> Table:
    """Get a valid PixelTable view for a given table.

    Args:
        table: The PixelTable table.
        minimal_schema: A dictionary representing the minimal expected schema
            (keys are column names, values are expected types).

    Returns:
        A valid PixelTable view for the given table.

    Raises:
        ValueError: If the table is missing expected keys, has excluded keys,
        or if the existing view does not point to the provided table.
    """

    if isinstance(table, Table):
        table_columns = set(table.columns())
        expected_columns = set(minimal_schema.keys())

        if missing := expected_columns.difference(table_columns):
            raise ValueError(f"The table is missing required keys: {missing}")

    else:
        create_pxt_dirs_for_path(table)
        created_table = pxt.create_table(
            table,
            schema=minimal_schema,
            primary_key=primary_key,
        )
        assert isinstance(created_table, Table)
        table = created_table

    return table


def make_batch_ready_for_table(
    batch: dict[str, Any],
    to_insert: list[str],
    index_key: str = "idx",
) -> list[dict[str, Any]]:
    """Unwrap a batch of data into a list of data ready to be included in the table.

    Args:
        table: The PixelTable table to include the batch into.
        batch: A dictionary representing a batch of data (each key corresponds to stacked information).
        to_insert: A list of keys from the batch to insert or update in the table.
        index_key: The key in the batch that represents the index for matching rows in the table.

    Returns:
        A list of dictionaries, each representing a row to be inserted or updated in the table.
    Raises:
        ValueError: If the index key is not found in the batch or if it is included in the to_insert list.
    """
    to_insert_list = []
    for batch_idx, idx_value in enumerate(batch[index_key]):
        sample_idx = (
            idx_value.item() if isinstance(idx_value, torch.Tensor) else idx_value
        )

        to_insert_dict = {
            key: (
                (
                    batch[key][batch_idx].item()
                    if batch[key][batch_idx].numel() == 1
                    else batch[key][batch_idx].detach().cpu().numpy()
                )
                if isinstance(batch[key][batch_idx], torch.Tensor)
                else batch[key][batch_idx]
            )
            for key in to_insert
        }
        to_insert_list.append(
            {
                index_key: sample_idx,
            }
            | to_insert_dict
        )

    return to_insert_list


def get_pxt_table_primary_keys(table: Table) -> set[str]:
    """Get the primary key columns of a PixelTable table.

    Args:
        table: The PixelTable table.

    Returns:
        A set of primary key column names.
    """
    return set(
        col_name
        for col_name, col_dict in table.get_metadata()["columns"].items()
        if col_dict.get("is_primary_key", False)
    )


def check_pxt_table_has_primary_key(table: Table, cols: set[str]) -> None:
    """Check if some columns are defined as primary keys in the PixelTable table.

    Args:
        table: The PixelTable table.
        cols: A set of column names to check.

    Raises:
        ValueError: If any of the specified columns are not primary keys in the table.
    """
    primary_keys = get_pxt_table_primary_keys(table)
    if missing := cols.difference(primary_keys):
        raise ValueError(
            f"The table does not have all specified columns as primary keys, "
            f"missing: {missing}"
        )


def get_max_value_in_column(
    table: Table,
    col_expr: Expr,
) -> int:
    """Get the maximum value in a specified column of a PixelTable table.

    Args:
        table: The PixelTable table.
        col_expr: The column expression to analyze.

    Returns:
        The maximum value in the specified column.

    Raises:
        ValueError: If the maximum value is not an integer.
    """
    if table.count() == 0:
        raise ValueError("The table is empty")

    max_value = [row["max"] for row in table.select(pxtf.max(col_expr)).collect()][0]  # type: ignore[call-arg]

    if not isinstance(max_value, int):
        raise TypeError("The maximum value is not an integer.")

    return max_value
